unit U8085;

interface
uses
 Classes,SysUtils,Windows,ULex;
type
 TGarbage=array[0..1023] of pointer;
 PGarbage=^TGarbage;

 TMemory=array[0..65535] of byte;
 PMemory=^TMemory;

 TSymbolInfo=record
  name:array[0..4] of char;
  code:word;
  sinfo:integer;
 end;

 TOPInfo=packed record
  id:byte;
  name:array[0..9] of char;
  size:byte;
 end;
 POPInfo=^TOPInfo;

 TInstruction=record
  opcode,lno,bno:integer;
  otype1,otype2,val1,val2:integer;
 end;
 PInstruction=^TInstruction;

 PSymbolNode=^TSymbolNode;
 TSymbolNode=record
  name:pchar;
  itype,info1,info2:integer;
  left,right,parent:PSymbolNode;
 end;

type
 TSymbolTable=class
  private
   froot:PSymbolNode;
  protected
   function GetSymbol(aname:pchar):PSymbolNode;
  public
   function AddSymbol(aname:pchar;atype,ainfo,bno:integer):boolean;
   function FindSymbol(aname:pchar):PSymbolNode;
   procedure Clear;
   destructor Destroy;override;
  end;

 T8085 = class
  private
   fcall,fret:boolean;
   fstart:boolean;flda:integer;
   frega,fregb,fregc:byte;
   fregd,frege,fregh:byte;
   fregl,fregpsw:byte;
   frec,fregpc,fregsp:word;
   fgcount,ftokencnt:integer;
   ftokens:array[0..14] of TToken;
   fgarbage:PGarbage;
   fstable:TSymbolTable;
   fmemory:PMemory;
   finstrcts:TList;
   flex:TLex;
   ferrline,ferrtype:integer;
   foptable:pointer;
  protected
   procedure ICall(add1:word);
   procedure IRet;
   procedure IAdd(op1:byte;op2:byte);
   procedure ISub(op1:byte;op2:byte);
   procedure IAnd(op1:byte);
   procedure IOr(op1:byte);
   procedure IXor(op1:byte);
   procedure ICmp(op1:byte);
   function  IInr(op:byte;delta:integer):byte;
   procedure LoadInstructions;
   function IsOpCode(asymbol:Pchar):integer;
   function IsRegister(asymbol:Pchar):integer;
   function IsRegister8(acode:integer):boolean;
   function IsRegister16(acode:integer):boolean;
   function ParseTokens(curline:PString;lno:integer;var bno:integer;sline:boolean=false):integer;
   function GetInstructionFormat(sline:boolean=false):integer;
   function GetOPInfo(opcode:byte):POPInfo;
   function IsValidLabel(asymbol:PChar):boolean;
   function IsValidInstruction(op,ond1,ond2:integer):boolean;
   function ExtractSymbol(aline:Pstring;tkninfo:PToken):Pchar;
   function GetOperandType(acode:integer;asymbol:PChar):integer;
   procedure ClearGarbage;
   procedure ClearInstructions;
   function WriteInstruction(ains:PInstruction;var bno:integer;var lda:integer):boolean;
   function WriteCode:integer;
   function GetInsCount:integer;
   function FGetSign:integer;
   function FGetParity:integer;
   function FGetZero:integer;
   function FGetCarry:integer;
   function FGetAuxCarry:integer;
   procedure FSetSign(data:integer);
   procedure FSetParity(data:integer);
   procedure FSetZero(data:integer);
   procedure FSetCarry(data:integer);
   procedure FSetAuxCarry(data:integer);
   function GetRegM:byte;
   procedure SetRegM(data:byte);
  public
   constructor Create;
   destructor Destroy;override;
   procedure ExecuteLine;
   procedure Terminate;
   function GetInstruction(amem:word;var asize:integer;var anext:integer):string;
   function CompileLine(aline:string;pos:integer):boolean;
   function CompileCode(alines:TStrings):integer;
   property RegisterA:byte read frega write frega;
   property RegisterB:byte read fregb write fregb;
   property RegisterC:byte read fregc write fregc;
   property RegisterD:byte read fregd write fregd;
   property RegisterE:byte read frege write frege;
   property RegisterH:byte read fregh write fregh;
   property RegisterL:byte read fregl write fregl;
   property RegisterM:byte read GetRegM write SetRegM;
   property RegisterPSW:byte read fregpsw write fregpsw;
   property RegisterPC:word read fregpc write fregpc;
   property RegisterSP:word read fregsp write fregsp;
   property FSign:integer read FGetSign write FSetSign;
   property FParity:integer read FGetParity write FSetParity;
   property FZero:integer read FGetZero write FSetZero;
   property FCarry:integer read FGetCarry write FSetCarry;
   property FAuxCarry:integer read FGetAuxCarry write FSetAuxCarry;
   property ErrorLine:integer read ferrline;
   property ErrorType:integer read ferrtype;
   property InstructionCount:integer read GetInsCount;
   property Ready:boolean read fstart write fstart;
   property LoadAddress:integer read flda;
   property Memory:PMemory read FMemory;
   property ProcCalled:boolean read fcall;
   property ProcReturned:boolean read fret;
   property ProcIndex:word read frec; 
  end;

const ERR_NFSYMB =$02;
const ERR_MSYMB  =$03;
const ERR_ARG    =$04;
const ERR_LRGE   =$05;
const ERR_PARSE1 =$08;
const ERR_INVLAB =$10;
const ERR_OPCODE =$11;
const ACC_LINE   =-1;

procedure DecToHex(data:word;s:pchar);
function HexToDec(c:char):integer;overload;
function HexToDec(s:pchar):integer;overload;

implementation

{$R optable.res}

function GetParity(a:byte):integer;forward;
const Hex1:array['0'..'9'] of integer=(0,1,2,3,4,5,6,7,8,9);
const Hex2:array['A'..'F'] of integer=(10,11,12,13,14,15);
const Dec :array[0..15] of char=('0','1','2','3','4','5','6','7','8',
                                 '9','A','B','C','D','E','F');

const OPD_LABEL  =$1;
const OPD_REG8   =$2;
const OPD_REG16  =$4;
const OPD_VALUE  =$8;

const OP_DW     =$510;
const OP_DB     =$520;
const OP_DSN    =$530;
const OP_DS     =$540;
const OP_EQU    =$550;
const OP_ORG    =$560;

const MSK_OP1   =$FFF0000;
const MSK_OP2   =$000FFF0;
const OP_MAX    =86;
const REG_MIN   =1;
const REG_MAX   =10;

const TOK_NUL       =$0000;
const TOK_CMT       =$0001;
const TOK_IGN       =$00FF;

const TOK_BYT       =$0101;
const TOK_WRD       =$0102;
const TOK_STR       =$0104;

const TOK_CLN       =$0201;
const TOK_SEM       =$0202;
const TOK_CMA       =$0204;

const TOK_DFB       =$0301;
const TOK_DFW       =$0302;
const TOK_DFS       =$0304;
const TOK_EQU       =$0308;
const TOK_ORG       =$0310;

const TOK_LAB       =$0401;
const TOK_SYM       =$0402;

const TOK_NUM       =TOK_WRD or TOK_BYT;

const INS_MAXN      =22;
const INS_MAXL      =06;

const InsFormat:array[0..22,0..6] of integer =
 (
  {00}(TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN),

  {01}(TOK_LAB,TOK_CLN,TOK_SYM,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN),
  {02}(TOK_LAB,TOK_CLN,TOK_SYM,TOK_SYM,TOK_IGN,TOK_IGN,TOK_IGN),
  {03}(TOK_LAB,TOK_CLN,TOK_SYM,TOK_NUM,TOK_IGN,TOK_IGN,TOK_IGN),

  {04}(TOK_LAB,TOK_CLN,TOK_SYM,TOK_SYM,TOK_CMA,TOK_SYM,TOK_IGN),
  {05}(TOK_LAB,TOK_CLN,TOK_SYM,TOK_SYM,TOK_CMA,TOK_NUM,TOK_IGN),
  {06}(TOK_LAB,TOK_CLN,TOK_SYM,TOK_NUM,TOK_CMA,TOK_SYM,TOK_IGN),

  {07}(TOK_SYM,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN),

  {08}(TOK_SYM,TOK_SYM,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN),
  {09}(TOK_SYM,TOK_NUM,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN),

  {10}(TOK_SYM,TOK_SYM,TOK_CMA,TOK_SYM,TOK_IGN,TOK_IGN,TOK_IGN),
  {11}(TOK_SYM,TOK_SYM,TOK_CMA,TOK_NUM,TOK_IGN,TOK_IGN,TOK_IGN),
  {12}(TOK_SYM,TOK_NUM,TOK_CMA,TOK_SYM,TOK_IGN,TOK_IGN,TOK_IGN),

  {13}(TOK_SYM,TOK_DFS,TOK_NUM,TOK_CMT,TOK_IGN,TOK_IGN,TOK_IGN),
  {14}(TOK_SYM,TOK_DFS,TOK_STR,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN),
  {15}(TOK_SYM,TOK_DFW,TOK_NUM,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN),
  {16}(TOK_SYM,TOK_DFB,TOK_BYT,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN),

  {17}(TOK_SYM,TOK_EQU,TOK_NUM,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN),
  {18}(TOK_ORG,TOK_NUM,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN),

  {19}(TOK_DFS,TOK_NUM,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN),
  {20}(TOK_DFS,TOK_STR,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN),
  {21}(TOK_DFW,TOK_NUM,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN),
  {22}(TOK_DFB,TOK_BYT,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN,TOK_IGN)
 );

const RegFormat:array[1..10] of TSymbolInfo=
 (
  (name:'A';  code:$0008;sinfo:$800),
  (name:'B';  code:$0001;sinfo:$400),
  (name:'C';  code:$0002;sinfo:$200),
  (name:'D';  code:$1003;sinfo:$100),
  (name:'E';  code:$0004;sinfo:$80),
  (name:'H';  code:$2005;sinfo:$40),
  (name:'L';  code:$0006;sinfo:$20),
  (name:'M';  code:$0007;sinfo:$10),
  (name:'PSW';code:$3000;sinfo:$4),
  (name:'SP'; code:$3001;sinfo:$8)
 );

const OpFormat:array[0..86] of TSymbolInfo=
 (
  (name:'ACI'; code:$CE;sinfo:$0020002),
  (name:'ADC'; code:$87;sinfo:$FF00001),
  (name:'ADD'; code:$7F;sinfo:$FF00001),
  (name:'ADI'; code:$C6;sinfo:$0020002),
  (name:'ANA'; code:$9F;sinfo:$FF00001),
  (name:'ANI'; code:$E6;sinfo:$0020002),
  (name:'CALL';code:$CD;sinfo:$0010003),
  (name:'CC';  code:$DC;sinfo:$0010003),
  (name:'CM';  code:$FC;sinfo:$0010003),
  (name:'CMA'; code:$2F;sinfo:$0000001),
  (name:'CMC'; code:$3F;sinfo:$FE0FF01),
  (name:'CMP'; code:$B7;sinfo:$FF00001),
  (name:'CNC'; code:$D4;sinfo:$0010003),
  (name:'CNZ'; code:$C4;sinfo:$0010003),
  (name:'CP';  code:$F4;sinfo:$0010003),
  (name:'CPE'; code:$EC;sinfo:$0010003),
  (name:'CPI'; code:$FE;sinfo:$0020002),
  (name:'CPO'; code:$E4;sinfo:$0010003),
  (name:'CZ';  code:$CC;sinfo:$0010003),
  (name:'DAA'; code:$27;sinfo:$0000001),
  (name:'DAD'; code:$09;sinfo:$0000001),
  (name:'DCR'; code:$05;sinfo:$FF00001),
  (name:'DCX'; code:$0B;sinfo:$5000001),
  (name:'DI';  code:$F3;sinfo:$0000001),
  (name:'EI';  code:$FB;sinfo:$0020001),
  (name:'HLT'; code:$76;sinfo:$0000001),
  (name:'IN';  code:$DB;sinfo:$0020002),
  (name:'INR'; code:$04;sinfo:$FF00001),
  (name:'INX'; code:$03;sinfo:$5480001),
  (name:'JC';  code:$DA;sinfo:$0010003),
  (name:'JM';  code:$FA;sinfo:$0010003),
  (name:'JMP'; code:$C3;sinfo:$0010003),
  (name:'JNC'; code:$D2;sinfo:$0010003),
  (name:'JNZ'; code:$C2;sinfo:$0010003),
  (name:'JP';  code:$F2;sinfo:$0010003),
  (name:'JPE'; code:$EA;sinfo:$0010003),
  (name:'JPO'; code:$E2;sinfo:$0010003),
  (name:'JZ';  code:$CA;sinfo:$0010003),
  (name:'LDA'; code:$3A;sinfo:$0010003),
  (name:'LDAX';code:$0A;sinfo:$5480001),
  (name:'LHLD';code:$2A;sinfo:$0000003),
  (name:'LXI'; code:$01;sinfo:$5480003),
  (name:'MOV'; code:$40;sinfo:$010FE01),
  (name:'MVI'; code:$06;sinfo:$FF00022),
  (name:'NOP'; code:$00;sinfo:$0000001),
  (name:'ORA'; code:$AF;sinfo:$FF00001),
  (name:'ORI'; code:$F6;sinfo:$0020002),
  (name:'OUT'; code:$D3;sinfo:$0020002),
  (name:'PCHL';code:$E9;sinfo:$0000001),
  (name:'POP'; code:$C1;sinfo:$5440001),
  (name:'PUSH';code:$C5;sinfo:$5440001),
  (name:'RAL'; code:$17;sinfo:$0000001),
  (name:'RAR'; code:$1F;sinfo:$0000001),
  (name:'RC';  code:$D8;sinfo:$0000001),
  (name:'RET'; code:$C9;sinfo:$0000001),
  (name:'RIM'; code:$20;sinfo:$0000001),
  (name:'RLC'; code:$07;sinfo:$0000001),
  (name:'RM';  code:$F8;sinfo:$0000001),
  (name:'RNC'; code:$D0;sinfo:$0000001),
  (name:'RNZ'; code:$C0;sinfo:$0000001),
  (name:'RP';  code:$F0;sinfo:$0000001),
  (name:'RPE'; code:$E8;sinfo:$0000001),
  (name:'RPO'; code:$E0;sinfo:$0000001),
  (name:'RRC'; code:$0F;sinfo:$5480001),
  (name:'RST0';code:$C7;sinfo:$0000001),
  (name:'RST1';code:$CF;sinfo:$0000001),
  (name:'RST2';code:$D7;sinfo:$0000001),
  (name:'RST3';code:$DF;sinfo:$0000001),
  (name:'RST4';code:$E7;sinfo:$0000001),
  (name:'RST5';code:$EF;sinfo:$0000001),
  (name:'RST6';code:$F7;sinfo:$0000001),
  (name:'RST7';code:$FF;sinfo:$0000001),
  (name:'RZ';  code:$C8;sinfo:$0000001),
  (name:'SBB'; code:$97;sinfo:$FF00001),
  (name:'SBI'; code:$DE;sinfo:$0020002),
  (name:'SHLD';code:$22;sinfo:$0000003),
  (name:'SIM'; code:$30;sinfo:$0010001),
  (name:'SPHL';code:$F9;sinfo:$0000001),
  (name:'STA'; code:$32;sinfo:$0000003),
  (name:'STAX';code:$02;sinfo:$5000001),
  (name:'STC'; code:$37;sinfo:$0010001),
  (name:'SUB'; code:$8F;sinfo:$FF00001),
  (name:'SUI'; code:$D6;sinfo:$0020002),
  (name:'XCHG';code:$EB;sinfo:$0000001),
  (name:'XRA'; code:$A7;sinfo:$FF00001),
  (name:'XRI'; code:$EE;sinfo:$0020002),
  (name:'XTHL';code:$E3;sinfo:$0000001)
 );

function TSymbolTable.AddSymbol(aname:pchar;atype,ainfo,bno:integer):boolean;
var
 pntnode,tmpnode:PSymbolNode;
 res:integer;
begin
 result:=false;
 if(froot=nil) then
  begin
   new(froot);
   froot.left:=nil;froot.right:=nil;froot.parent:=nil;
   froot.name:=aname;
   froot.info1:=ainfo;
   froot.itype:=atype;
   froot.info2:=bno;
   result:=true; exit;
  end;
 pntnode:=GetSymbol(aname);
 res:=stricomp(aname,pntnode.name);
 if(res=0) then exit;
 new(tmpnode);
 tmpnode.name:=aname;
 tmpnode.left:=nil;
 tmpnode.right:=nil;
 tmpnode.info1:=ainfo;
 tmpnode.info2:=bno;
 tmpnode.itype:=atype;
 tmpnode.parent:=pntnode;
 if(res<0) then pntnode.left:=tmpnode
 else pntnode.right:=tmpnode;
 result:=true;
end;

function TSymbolTable.GetSymbol(aname:pchar):PSymbolNode;
var
 curnode:PSymbolNode;
 res:integer;
begin
 result:=nil;
 curnode:=froot;
 while(curnode<>nil) do
  begin
   res:=stricomp(aname,curnode.name);
   if(res=0) then
    begin
     result:=curnode;
     exit;
    end
   else if res<0 then
    begin
     if(curnode.left=nil) then
      begin
       result:=curnode;
       exit;
      end;
     curnode:=curnode.left;
    end
   else
    begin
     if(curnode.right=nil) then
      begin
       result:=curnode;
       exit;
      end;
     curnode:=curnode.right;
    end;
  end;
end;

function TSymbolTable.FindSymbol(aname:pchar):PSymbolNode;
var
 tmpnode:PSymbolNode;
begin
 tmpnode:=GetSymbol(aname);
 if (tmpnode=nil) or (stricomp(tmpnode.name,aname)<>0) then result:=nil
 else result:=tmpnode;
end;

procedure TSymbolTable.Clear;
 procedure Remove(anode:PSymbolNode);
 begin
  if(anode=nil) then exit;
  Remove(anode.left);
  Remove(anode.right);
  dispose(anode);
 end;
begin
 Remove(froot);
 froot:=nil;
end;

destructor TSymbolTable.Destroy;
begin
 Clear;
 inherited;
end;

constructor T8085.Create;
begin
 new(fmemory);new(fgarbage);
 flex:=TLex.Create;
 finstrcts:=TList.Create;
 flda:=$0040;fregpc:=$0040;fregsp:=$FF00;
 fstable:=TSymbolTable.Create;
 LoadInstructions;
end;

procedure T8085.ClearGarbage;
var
 i:integer;
begin
 for i:=0 to fgcount-1 do
  freemem(fgarbage[i]);
 fgcount:=0;
end;

function T8085.GetInsCount:integer;
begin
 result:=finstrcts.Count;
end;

procedure T8085.FSetSign(data:integer);
begin
 if data=0 then fregpsw:=fregpsw and not $80
 else if data=1 then fregpsw:=fregpsw or $80;
end;

procedure T8085.FSetParity(data:integer);
begin
 if data=0 then fregpsw:=fregpsw and not $4
 else if data=1 then fregpsw:=fregpsw or $4;
end;

procedure T8085.FSetZero(data:integer);
begin
 if data=0 then fregpsw:=fregpsw and not $40
 else if data=1 then fregpsw:=fregpsw or $40;
end;

procedure T8085.FSetCarry(data:integer);
begin
 if data=0 then fregpsw:=fregpsw and not $1
 else if data=1 then fregpsw:=fregpsw or $1;
end;

procedure T8085.FSetAuxCarry(data:integer);
begin
 if data=0 then fregpsw:=fregpsw and not $10
 else if data=1 then fregpsw:=fregpsw or $10;
end;

function T8085.FGetSign:integer;
begin
 result:=fregpsw shr 7 and $1;
end;

function T8085.FGetParity:integer;
begin
 result:=fregpsw shr 2 and $1;
end;

function T8085.FGetZero:integer;
begin
 result:=fregpsw shr 6 and $1;
end;

function T8085.FGetCarry:integer;
begin
 result:=fregpsw and $1;
end;

function T8085.FGetAuxCarry:integer;
begin
 result:=fregpsw shr 4 and $1;
end;

function T8085.GetRegM:byte;
begin
 result:=fmemory[fregh shl 8 or fregl];
end;

procedure T8085.SetRegM(data:byte);
begin
 fmemory[fregh shl 8 or fregl]:=data;
end;

procedure T8085.ClearInstructions;
var
 i:integer;
begin
 for i:=0 to finstrcts.count-1 do dispose(finstrcts[i]);
 finstrcts.Clear;
end;

function T8085.GetInstruction(amem:word;var asize:integer;var anext:integer):string;
var
 opinfo:POpInfo;
 d1:word;
 val:array[0..3] of char;
begin
 opinfo:=GetOpInfo(fmemory[amem]);
 result:=opinfo.name;asize:=opinfo.size;
 if(asize>1) then
  begin
   if asize=2 then
    begin
     d1:=fmemory[amem+1];
     DecToHex(d1,val);
     result:=result + ' ' +val[2]+val[3];
    end
   else if asize=3 then
    begin
     d1:=fmemory[amem+2] shl 8 or fmemory[amem+1];
     DecToHex(d1,val);
     result:=result + ' ' +val;
    end;
  end;
 anext:=amem+asize;
end;

procedure T8085.Terminate;
begin
 ClearInstructions;ClearGarbage;
 fstable.Clear;fstart:=false;
 frega:=0;fregb:=0;fregc:=0;fregd:=0;frege:=0;fregh:=0;
 fregl:=0;flda:=$0040;fregpc:=$0040;fregsp:=$FF00;
 fregpsw:=0;frec:=0;fcall:=false;fret:=false;
end;

destructor T8085.Destroy;
begin
 dispose(fmemory);
 ClearInstructions;
 freeandnil(finstrcts);
 freeandnil(flex);
 freeandnil(fstable);
 ClearGarbage;
 dispose(fgarbage);
 inherited;
end;

procedure T8085.LoadInstructions;
var
 h1,h2:integer;
begin
 flex.AddToken('[+:D,d][+:B,b]',TOK_DFB);
 flex.AddToken('[+:D,d][+:W,w]',TOK_DFW);
 flex.AddToken('[+:D,d][+:S,s]',TOK_DFS);
 flex.AddToken('[+:O,o][+:R,r][+:G,g]',TOK_ORG);
 flex.AddToken('[+:E,e][+:Q,q][+:U,u]',TOK_EQU);
 flex.AddToken('"([-:"]|"")*"',TOK_STR);
 flex.AddToken(';\v*',TOK_CMT);
 flex.AddToken(',',TOK_CMA);
 flex.AddToken(':',TOK_CLN);
 flex.AddToken('0*(\x[+:H,h]|\x\x[+:H,h])',TOK_BYT,1,true);
 flex.AddToken('0*(\x\x\x[+:H,h]|\x\x\x\x[+:H,h])',TOK_WRD,1,true);
 flex.AddToken('\a(\a|\d)*:\p',TOK_LAB,2);
 flex.AddToken('\a(\a|\d)*',TOK_SYM);
 h1:=FindResource(0,'#102','OP_TABLE');
 h2:=LoadResource(0,h1);
 foptable:=LockResource(h2);
end;

function T8085.ExtractSymbol(aline:PString;tkninfo:PToken):Pchar;
var
 i,j,len:integer;
begin
 result:=nil;
 if(aline=nil) or (tkninfo=nil) then exit;
 j:=0;len:=tkninfo.stop-tkninfo.start+1;
 Getmem(result,len+1);
 for i:=tkninfo.start to tkninfo.stop do
  begin
   result[j]:=aline^[i];
   j:=j+1;
  end;
 result[len]:=#0;
 fgarbage[fgcount]:=result;
 fgcount:=fgcount+1;

end;

function T8085.IsRegister8(acode:integer):boolean;
begin
 result:=false;
 if (acode>=1) and (acode<=8) then result:=true;
end;

function T8085.IsRegister16(acode:integer):boolean;
begin
 result:=false;
 if acode in [2,4,6,9,10] then result:=true;
end;

function T8085.IsRegister(asymbol:Pchar):integer;
var
 l,h,m:integer;
 res:integer;
begin
 result:=-1;
 l:=REG_MIN;h:=REG_MAX;m:=(l+h) div 2;
 while(l<=h) do
  begin
   res:=stricomp(asymbol,RegFormat[m].name);
   if(res=0) then
    begin
     result:=m;
     exit;
    end
   else if( res<0) then h:=m-1
   else l:=m+1;
   m:=(l+h) div 2;
  end;
end;

function T8085.IsOpCode(asymbol:Pchar):integer;
var
 l,m,h:integer;
 res:integer;
begin
 result:=-1;
 l:=0;h:=OP_MAX;m:=(l+h) div 2;
 while(l<=h) do
  begin
   res:=stricomp(asymbol,opformat[m].name);
   if(res=0) then
    begin
     result:=m;
     exit;
    end
   else if(res<0) then h:=m-1
   else l:=m+1;
   m:=(l+h) div 2;
  end;
end;

function T8085.IsValidLabel(asymbol:PChar):boolean;
begin
 result:=false;
 if(asymbol=nil) then exit;
 if IsRegister(asymbol)>=0 then exit;
 if IsOpCode(asymbol)>=0 then exit;
 result:=true;
end;

function T8085.GetOPInfo(opcode:byte):POpInfo;
begin
 result:=POpInfo(longword(foptable)+opcode*sizeof(TOPInfo));
end;

function T8085.GetInstructionFormat(sline:boolean=false):integer;
var
 i,j:integer;
 fg:boolean;
begin
 result:=-1;
 if ftokencnt>INS_MAXL+1 then exit;
 if ftokencnt=0 then
  begin
   result:=0; exit;
  end;
 for i:=0 to INS_MAXN do
  begin
   fg:=true;
   for j:=0 to ftokencnt-1 do
    begin
     if (InsFormat[i,j] and ftokens[j].code)<>ftokens[j].code then
      begin
       fg:=false;break;
      end;
    end;
   if fg then
    begin
     if (ftokencnt<=INS_MAXL) and (InsFormat[i,ftokencnt]<>TOK_IGN) then continue;
     result:=i; break;
    end;
  end;
 if(sline)and(result>=0) then
  if not result in [23,27,28,29,35,36,37,38,39] then result:=-1;
end;

function T8085.IsValidInstruction(op,ond1,ond2:integer):boolean;
var
 t1,t2:integer;
begin
 result:=false;
 if(op<0) then exit;
 ond1:=ond1 shl 12; ond2:=ond2 shl 4;
 t1:=OpFormat[op].sinfo and MSK_OP1;
 t2:=OpFormat[op].sinfo and MSK_OP2;
 if(t1>0) and (ond1=0) or ((t1=0) and (ond1>0)) then exit;
 if(t2>0) and (ond2=0) or ((t2=0) and (ond2>0)) then exit;
 if (t1 and ond1)<>ond1 then exit;
 if (t2 and ond2)<>ond2 then exit;
 result:=true;
end;

function T8085.GetOperandType(acode:integer;asymbol:PChar):integer;
begin
 result:=-1; if(asymbol=nil) then exit;
 if acode=TOK_WRD then result:=TOK_WRD
 else if acode=TOK_BYT then result:=TOK_BYT
 else if acode=TOK_SYM then
  begin
   result:=IsRegister(asymbol);if(result>=0) then exit;
   if IsValidLabel(asymbol) then result:=TOK_LAB;
  end;
end;


function T8085.WriteInstruction(ains:PInstruction;var bno:integer;var lda:integer):boolean;
var
 i,ccode,opcode:integer;
 reg1,reg2:integer;
 val:word;str:pchar;
 symbnode:PSymbolNode;
begin
 val:=0;result:=false;
 if(ains=nil) or (bno<0) or (bno>$FFFF) then
  begin
   ferrtype:=ERR_LRGE;exit;
  end;
 if ains.opcode=OP_ORG then
  begin
   bno:=ains.val1;
   result:=true;exit;
  end;
 if ains.opcode=OP_DB then
  begin
   fmemory[bno]:=(ains.val1) and $FF;
   bno:=bno+1;
   result:=true;exit;
  end;
 if ains.opcode=OP_DW then
  begin
   fmemory[bno]:=(ains.val1) and $FF;
   bno:=bno+1;
   fmemory[bno]:=(ains.val1 shr 8) and $FF;
   bno:=bno+1;
   result:=true;exit;
  end;
 if ains.opcode=OP_DSN then
  begin
   if bno+ains.val1>$FFFF then
    begin
     ferrtype:=ERR_LRGE;exit;
    end;
   bno:=bno+ains.val1;result:=true;exit;
  end;
 if ains.opcode=OP_DS then
  begin
   str:=pchar(ains.val1);inc(str);
   if bno+ains.val2>$FFFF then
    begin
     ferrtype:=ERR_LRGE;exit;
    end;
   for i:=1 to ains.val2 do
    begin
     fmemory[bno]:=byte(str^);inc(str);
     bno:=bno+1;
    end;
   result:=true;exit;
  end;
 if ains.opcode=OP_EQU then
  begin
   result:=true;exit;
  end;
 if (ains.opcode<0) or (ains.opcode>OP_MAX) then exit;
 opcode:=OpFormat[ains.opcode].code;lda:=bno;
 if bno+(OpFormat[ains.opcode].sinfo and $F)-1>$FFFF then
  begin
   ferrtype:=ERR_LRGE;exit;
  end; 
 if opcode in [$04,$05,$7F,$87,$8F,$97,$9F,$A7,$AF,$B7] then
  begin
   if (ains.otype1=0) or (ains.otype2<>0) then
    begin
     ferrtype:=ERR_ARG; 
     exit;
    end;
   ccode:=(OpFormat[ains.opcode].sinfo shr 16) and $FFF;
   reg1:=RegFormat[ains.val1].sinfo;
   if (reg1 and ccode)<>reg1 then
    begin
     ferrtype:=ERR_ARG;exit;
    end;
   reg1:=RegFormat[ains.val1].code and $FF;
   if opcode in [$04,$05] then opcode:=opcode+(reg1-1)*8
   else opcode:=opcode+reg1;
   fmemory[bno]:=opcode;bno:=bno+1;result:=true;
  end
 else if opcode=$40 then
  begin
   if (ains.otype1 and OPD_REG8>0) and (ains.otype2 and OPD_REG8>0) then
    begin
     reg1:=RegFormat[ains.val1].code and $FF;
     reg2:=RegFormat[ains.val2].code and $FF;
     if (reg1=$7) and (reg2=$7) then
      begin
       ferrtype:=ERR_ARG;exit;
      end; 
     opcode:=opcode+(reg2-1)+(reg1-1)*8;
     fmemory[bno]:=opcode;bno:=bno+1;result:=true;
    end
   else
    begin
     ferrtype:=ERR_ARG;exit;
    end;
  end
 else if opcode in [$02,$03,$09,$0A,$0B,$C1,$C5] then
  begin
   if(ains.otype1 and OPD_REG16=0) or (ains.otype2<>0) then
    begin
     ferrtype:=ERR_ARG;exit;
    end;
   reg1:=(RegFormat[ains.val1].code shr 8) and $FF;
   if (opcode in [$02,$0A]) and (ains.val1 in [6,9,10])  then
    begin
     ferrtype:=ERR_ARG;exit;
    end;
   if (opcode in [$03,$09,$0B]) and (ains.val1=9) then
    begin
     ferrtype:=ERR_ARG;exit;
    end;
   if (opcode in [$C1,$C5]) and (ains.val1=10) then
    begin
     ferrtype:=ERR_ARG;exit;
    end;
   opcode:=opcode+reg1;
   fmemory[bno]:=opcode;bno:=bno+1;result:=true;exit;
  end
 else if opcode in  [$00,$07,$17,$20,$27,$30,$37,$76,$0F,$1F,
                     $2F,$3F,$C0,$C7,$C8,$C9,$CF,$D0,$D7,$D8,
                     $DF,$E0,$E3,$E7,$E8,$E9,$EB,$EF,$F0,$F3,
                     $F7,$F8,$F9,$FB,$FF] then
  begin
   if(ains.otype1<>0) or (ains.otype2<>0) then
    begin
     ferrtype:=ERR_ARG;exit;
    end;
   fmemory[bno]:=opcode;bno:=bno+1;result:=true;exit;
  end
 else if opcode in [$C6,$CE,$D3,$D6,$DB,$DE,$E6,$EE,$F6,$FE] then
  begin
   if (ains.otype2<>0) or (ains.otype2 and (OPD_REG8 or OPD_REG16)>0) then
    begin
     ferrtype:=ERR_ARG;exit;
    end; 
   if ains.otype1=OPD_VALUE then val:=ains.val1
   else if ains.otype1=OPD_LABEL then
    begin
     symbnode:=fstable.FindSymbol(pchar(ains.val1));
     if (symbnode=nil) or (symbnode.itype<>OP_EQU) then
      begin
       ferrtype:=ERR_NFSYMB;exit;
      end;
     val:=symbnode.info2;
    end;
   if (val>$FF) then
    begin
     ferrtype:=ERR_ARG;exit;
    end;
   fmemory[bno]:=opcode;bno:=bno+1;
   fmemory[bno]:=val and $FF;bno:=bno+1;
   result:=true;
  end
 else if opcode in [$22,$2A,$32,$3A,$C2,$C3,$C4,$CA,$CC,$CD,$D2,
                    $D4,$DA,$DC,$E2,$E4,$EA,$EC,$F2,$F4,$FA,$FC] then
  begin
   if (ains.otype2<>0) or (ains.otype1 and (OPD_REG8 or OPD_REG16)>0) then
    begin
     ferrtype:=ERR_ARG;exit;
    end;
   if ains.otype1=OPD_VALUE then val:=ains.val1
   else if ains.otype1=OPD_LABEL then
    begin
     symbnode:=fstable.FindSymbol(pchar(ains.val1));
     if (symbnode=nil) then
      begin
       ferrtype:=ERR_NFSYMB;exit;
      end;
     val:=symbnode.info2;
    end;
   fmemory[bno]:=opcode;bno:=bno+1;
   fmemory[bno]:=val and $FF;bno:=bno+1;
   fmemory[bno]:=(val shr 8) and $FF;bno:=bno+1;
   result:=true;
  end
 else if opcode=$06 then
  begin
   if (ains.otype1 and OPD_REG8=0) or (ains.otype2 and (OPD_REG8 or OPD_REG16)>0) then
    begin
     ferrtype:=ERR_ARG;exit;
    end;
   if ains.otype2=OPD_VALUE then val:=ains.val2
   else if ains.otype2=OPD_LABEL then
    begin
     symbnode:=fstable.FindSymbol(pchar(ains.val2));
     if (symbnode=nil) then
      begin
       ferrtype:=ERR_NFSYMB;exit;
      end;
     if symbnode.itype=OP_EQU then val:=symbnode.info2
     else if (symbnode.itype=OP_DB) or (symbnode.itype=OP_DW) then
      begin
       ferrtype:=ERR_ARG;exit;
      end;
    end;
   if (val>$FF) then
    begin
     ferrtype:=ERR_ARG;exit;
    end;
   reg1:=RegFormat[ains.val1].code and $FF;
   opcode:=opcode+(reg1-1)*8;
   fmemory[bno]:=opcode;bno:=bno+1;
   fmemory[bno]:=val and $FF;bno:=bno+1;
   result:=true;
  end
 else if opcode=$01 then
  begin
   if (ains.otype1 and OPD_REG16=0) or (ains.otype2 and (OPD_REG8 or OPD_REG16)>0)
      or (ains.val1=$40) then
    begin
     ferrtype:=ERR_ARG;exit;
    end;
   if ains.otype2=OPD_VALUE then val:=ains.val2
   else if ains.otype2=OPD_LABEL then
    begin
     symbnode:=fstable.FindSymbol(pchar(ains.val2));
     if (symbnode=nil) then
      begin
       ferrtype:=ERR_NFSYMB;exit;
      end;
     val:=symbnode.info2;
    end;
   reg1:=(RegFormat[ains.val1].code shr 8) and $FF;
   opcode:=opcode+reg1;
   fmemory[bno]:=opcode;bno:=bno+1;
   fmemory[bno]:=val and $FF;bno:=bno+1;
   fmemory[bno]:=(val shr 8) and $FF;bno:=bno+1;
   result:=true;
  end;
end;

function T8085.WriteCode:integer;
var
 i,lda,bno:integer;
 curins:PInstruction;
begin
 bno:=fregpc;result:=-1;lda:=-1;flda:=-1;
 for i:=0 to finstrcts.count-1 do
  begin
   curins:=finstrcts[i];
   if not WriteInstruction(curins,bno,lda) then
    begin
     result:=curins.lno;
     exit;
    end;
   if (flda<0) and (lda>=0) then flda:=lda;
  end;
 if flda<0 then flda:=fregpc else fregpc:=flda;
end;

function T8085.ParseTokens(curline:PString;lno:integer;var bno:integer;sline:boolean=false):integer;
var
 idx,fno,res:integer;
 asymbol:pchar;
 curtoken:Ptoken;
 curIns:PInstruction;
begin
 result:=ERR_PARSE1; if curline=nil then exit;
 fno:=length(curline^); if(ftokencnt=0) and(fno>0) then exit;
 fno:=GetInstructionFormat(sline);if fno<0 then exit;
 if(fno=0) then
  begin
   result:=ACC_LINE;
   exit;
  end;
 new(curIns);ZeroMemory(curins,sizeof(TInstruction));
 curins.lno:=lno;curins.bno:=bno;
 if (fno>=1) and (fno<=6) then
  begin
   curtoken:=@ftokens[0];asymbol:=ExtractSymbol(curline,curtoken);
   if IsValidLabel(asymbol) and (fstable.FindSymbol(asymbol)=nil) then fstable.AddSymbol(asymbol,OP_EQU,lno,bno)
   else
    begin
     result:=ERR_MSYMB;
     exit;
    end;
   curtoken:=@ftokens[2];asymbol:=ExtractSymbol(curline,curtoken);
   res:=IsOpCode(asymbol);
   if res<0 then
    begin
     result:=ERR_OPCODE;
     dispose(curins); exit;
    end;
   curIns.opcode:=res;
  end
 else if (fno>=7) and (fno<=12) then
  begin
   curtoken:=@ftokens[0];asymbol:=ExtractSymbol(curline,curtoken);
   res:=IsOpCode(asymbol);
   if res<0 then
    begin
     result:=ERR_OPCODE;
     dispose(curins);exit;
    end;
   curIns.opcode:=res;
  end;
 if fno in [2,4,5,8,10,11] then
  begin
   if fno in [2,4,5] then idx:=3 else idx:=1;
   curtoken:=@ftokens[idx];asymbol:=ExtractSymbol(curline,curtoken);
   res:=GetOperandType(curtoken.code,asymbol);
   if IsRegister8(res) then
    begin
     curins.val1:=res;
     curins.otype1:=curins.otype1 or OPD_REG8;
    end;
   if IsRegister16(res) then
    begin
     curins.val1:=res;
     curins.otype1:=curins.otype1 or OPD_REG16;
    end
   else if res=TOK_LAB then
    begin
     curins.otype1:=OPD_LABEL;
     curins.val1:=integer(asymbol);
    end;
   if(res<0) then
    begin
     result:=ERR_ARG;
     dispose(curins);exit;
    end;
  end;
 if fno in [4,6,10,12] then
  begin
   if fno in [4,6] then idx:=5 else idx:=3;
   curtoken:=@ftokens[idx];asymbol:=ExtractSymbol(curline,curtoken);
   res:=GetOperandType(curtoken.code,asymbol);
   if IsRegister8(res) then
    begin
     curins.val2:=res;
     curins.otype2:=curins.otype2 or OPD_REG8;
    end;
   if IsRegister16(res) then
    begin
     curins.val2:=res;
     curins.otype2:=curins.otype2 or OPD_REG16;
    end
   else if res=TOK_LAB then
    begin
     curins.otype2:=OPD_LABEL;
     curins.val2:=integer(asymbol);
    end;
   if(res<0) then
    begin
     result:=ERR_ARG;
     dispose(curins);exit;
    end;
  end;
 if fno in [3,6,9,12,18] then
  begin
   if fno in [3,6] then idx:=3 else idx:=1;
   curtoken:=@ftokens[idx];asymbol:=ExtractSymbol(curline,curtoken);
   res:=HextoDec(asymbol);
   curins.otype1:=OPD_VALUE;
   curins.val1:=res;
  end;
 if fno in [5,11] then
  begin
   if fno=5 then idx:=5 else idx:=3;
   curtoken:=@ftokens[idx];asymbol:=ExtractSymbol(curline,curtoken);
   res:=HextoDec(asymbol);
   curins.otype2:=OPD_VALUE;
   curins.val2:=res;
  end;
 if (fno>=1) and (fno<=12) then result:=ACC_LINE
 else if (fno>=13) and (fno<=16) then
  begin
   curtoken:=@ftokens[0];asymbol:=ExtractSymbol(curline,curtoken);
   if not(IsValidLabel(asymbol)) then
    begin
     result:=ERR_INVLAB;
     dispose(curins);exit;
    end;
   if (fstable.FindSymbol(asymbol)<>nil) then
    begin
     result:=ERR_MSYMB;
     dispose(curins);
     exit;
    end;
   if fno=15 then
    begin
     fstable.AddSymbol(asymbol,OP_DW,lno,bno);
     curins.opcode:=OP_DW;
    end
   else if fno=16 then
    begin
     fstable.AddSymbol(asymbol,OP_DB,lno,bno);
     curins.opcode:=OP_DB;
    end
   else if fno=13 then
    begin
     fstable.AddSymbol(asymbol,OP_DSN,lno,bno);
     curins.opcode:=OP_DSN;
    end
   else
    begin
     fstable.AddSymbol(asymbol,OP_DS,lno,bno);
     curins.opcode:=OP_DS;
    end;
   curtoken:=@ftokens[2];asymbol:=ExtractSymbol(curline,curtoken);
   if fno=14 then
    begin
     curins.val1:=longword(asymbol);
     curins.val2:=strlen(asymbol)-2;
    end
   else curins.val1:=HextoDec(asymbol);
   result:=ACC_LINE;
  end
 else if(fno=17) then
  begin
   curtoken:=@ftokens[0];asymbol:=ExtractSymbol(curline,curtoken);
   curins.opcode:=OP_EQU;
   if not(IsValidLabel(asymbol)) then
    begin
     result:=ERR_INVLAB;
     dispose(curins);exit;
    end;
   if (fstable.FindSymbol(asymbol)<>nil) then
    begin
     result:=ERR_MSYMB;
     dispose(curins);
     exit;
    end;
   curtoken:=@ftokens[2];
   res:=HexToDec(ExtractSymbol(curline,curtoken));
   fstable.AddSymbol(asymbol,OP_EQU,lno,res);
   curins.val1:=res;result:=ACC_LINE;
  end
 else if (fno>=19) and (fno<=22) then
  begin
   curtoken:=@ftokens[1];asymbol:=ExtractSymbol(curline,curtoken);
   if fno=22 then curins.opcode:=OP_DB
   else if fno=21 then curins.opcode:=OP_DW
   else if fno=19 then curins.opcode:=OP_DSN
   else
    begin
     curins.opcode:=OP_DS;
     curins.val2:=strlen(asymbol)-2;
    end;  
   if fno=19 then
    begin
     curins.val1:=longword(asymbol);
     curins.val2:=strlen(asymbol)-2;
    end
   else curins.val1:=HextoDec(asymbol);
   result:=ACC_LINE;
  end
 else if fno=18 then
  begin
   curins.opcode:=OP_ORG;
   bno:=curins.val1;
   result:=ACC_LINE;
  end;
 if (curins.opcode>=0) and (curins.opcode<=OP_MAX) then
  bno:=bno+(OpFormat[curins.opcode].sinfo and $F)
 else if curins.opcode=OP_DW then bno:=bno+2
 else if curins.opcode=OP_DB then bno:=bno+1
 else if curins.opcode=OP_DSN then bno:=bno+curins.val1
 else if curins.opcode=OP_DS then bno:=bno+curins.val2; 
 if result=ACC_LINE then finstrcts.Add(curins);
end;

function T8085.CompileLine(aline:string;pos:integer):boolean;
begin
 ClearInstructions;result:=false;ferrtype:=0;ferrline:=-1;
 if length(trim(aline))=0 then
  begin
   ferrtype:=ERR_PARSE1;
   exit;
  end;
 ftokencnt:=flex.ListTokens(aline,ftokens,15);
 ferrtype:=ParseTokens(@aline,-1,pos,true);
 if(ferrtype>=0) then exit;
 pos:=PInstruction(finstrcts[0]).bno;
 if not WriteInstruction(PInstruction(finstrcts[0]),pos,pos) then exit;
 result:=true;
end;

function T8085.CompileCode(alines:TStrings):integer;
var
 i,bno:integer;
 aline:string;
begin
 bno:=fregpc;result:=-1;ClearInstructions;fstable.Clear;ClearGarbage;
 ferrtype:=0;ferrline:=-1;fstart:=false;
 if(alines=nil) then exit;
 for i:=0 to alines.Count-1 do
  begin
   aline:=trim(alines[i]);
   ftokencnt:=flex.ListTokens(aline,ftokens,15);
   ferrtype:=ParseTokens(@aline,i,bno);
   if (ferrtype>=0) then
    begin
     ferrline:=i;result:=i;
     exit;
    end;
  end;
 ferrline:=WriteCode;
 if ferrline<0 then  fstart:=true;
end;

function T8085.IInr(op:byte;delta:integer):byte;
var
 tmp1:integer;
begin
 tmp1:=op and $8;
 result:=op + delta;
 if result and $8<>tmp1 then fregpsw:=fregpsw or $10 else
  fregpsw:=fregpsw and not $10;
 if result=0 then fregpsw:=fregpsw or $40 else
  fregpsw:=fregpsw and not $40;
 if (result and $80)=0 then fregpsw:=fregpsw and not $80 else
  fregpsw:=fregpsw or $80;
 tmp1:=GetParity(result);
 if tmp1=1 then fregpsw:=fregpsw or $4 else fregpsw:=fregpsw and not $4;
end;

procedure T8085.IAdd(op1:byte;op2:byte);
var
 tmp1,tmp2:integer;
begin
 tmp1:=frega and $8;tmp2:=frega and $80;
 frega:=frega + op1 + op2;
 if (frega and $8=0) and (tmp1>0) then fregpsw:=fregpsw or $10 else
  fregpsw:=fregpsw and not $10;
 if (frega and $80=0) and (tmp2>0) then fregpsw:=fregpsw or $1 else
  fregpsw:=fregpsw and not $1;
 if frega=0 then fregpsw:=fregpsw or $40 else
  fregpsw:=fregpsw and not $40;
 if (frega and $80)=0 then fregpsw:=fregpsw and not $80 else
  fregpsw:=fregpsw or $80;
 tmp1:=GetParity(frega);
 if tmp1=1 then fregpsw:=fregpsw or $4 else fregpsw:=fregpsw and not $4;
end;

procedure T8085.ISub(op1:byte;op2:byte);
var
 tmp1,tmp2:integer;
begin
 tmp1:=frega and $8;tmp2:=frega and $80;
 frega:=frega - op1 - op2;
 if (frega and $8>0) and (tmp1=0) then fregpsw:=fregpsw or $10 else
  fregpsw:=fregpsw and not $10;
 if (frega and $80>0) and (tmp2=0) then fregpsw:=fregpsw or $1 else
  fregpsw:=fregpsw and not $1;
 if frega=0 then fregpsw:=fregpsw or $40 else
  fregpsw:=fregpsw and not $40;
 if (frega and $80)=0 then fregpsw:=fregpsw and not $80 else
  fregpsw:=fregpsw or $80;
 tmp1:=GetParity(frega);
 if tmp1=1 then fregpsw:=fregpsw or $4 else fregpsw:=fregpsw and not $4;
end;

procedure T8085.IAnd(op1:byte);
var
 tmp1:integer;
begin
 fregpsw:=fregpsw and not $1;
 fregpsw:=fregpsw and not $10;
 frega:=frega and op1;
 if frega=0 then fregpsw:=fregpsw or $40 else
  fregpsw:= fregpsw and not $40;
 if (frega and $80)=0 then fregpsw:=fregpsw and not $80 else
  fregpsw:=fregpsw or $80;
 tmp1:=GetParity(frega);
 if tmp1=1 then fregpsw:=fregpsw or $4 else fregpsw:=fregpsw and not $4;
end;

procedure T8085.IOr(op1:byte);
var
 tmp1:integer;
begin
 fregpsw:=fregpsw and not $1;
 fregpsw:=fregpsw and not $10;
 frega:=frega or op1;
 if frega=0 then fregpsw:=fregpsw or $40 else
  fregpsw:=fregpsw and not $40;
 if (frega and $80)=0 then fregpsw:=fregpsw and not $80 else
  fregpsw:=fregpsw or $80;
 tmp1:=GetParity(frega);
 if tmp1=1 then fregpsw:=fregpsw or $4 else fregpsw:=fregpsw and not $4;
end;

procedure T8085.IXor(op1:byte);
var
 tmp1:integer;
begin
 fregpsw:=fregpsw and not $1;
 fregpsw:=fregpsw and not $10;
 frega:=frega xor op1;
 if frega=0 then fregpsw:=fregpsw or $40 else
  fregpsw:=fregpsw and not $40;
 if (frega and $80)=0 then fregpsw:=fregpsw and not $80 else
  fregpsw:=fregpsw or $80;
 tmp1:=GetParity(frega);
 if tmp1=1 then fregpsw:=fregpsw or $4 else fregpsw:=fregpsw and not $4;
end;

procedure T8085.ICmp(op1:byte);
var
 tmp:integer;
begin
 tmp:=frega-op1;
 if tmp and $8<>0 then fregpsw:=fregpsw or $10 else
  fregpsw:=fregpsw and not $10;
 if tmp and $80<>0 then fregpsw:=fregpsw or $1 else
  fregpsw:=fregpsw and not $1;
 if tmp=0 then fregpsw:=fregpsw or $40 else
  fregpsw:=fregpsw and not $40;
 if (tmp and $80)=0 then fregpsw:=fregpsw and not $80 else
  fregpsw:=fregpsw or $80;
 tmp:=GetParity(tmp);
 if tmp=1 then fregpsw:=fregpsw or $4 else fregpsw:=fregpsw and not $4;
end;

procedure T8085.ICall(add1:word);
begin
 fcall:=true;frec:=frec+1;
 fregpc:=fregpc+3;
 fmemory[fregsp-1]:=(fregpc shr 8)and $FF;
 fmemory[fregsp-2]:=fregpc and $FF;
 fregsp:=fregsp-2;
 fregpc:=add1;
end;

procedure T8085.IRet;
begin
 fregpc:=fmemory[fregsp+1] shl 8 or fmemory[fregsp];
 fregsp:=fregsp+2;
 if frec>0 then frec:=frec-1;fret:=true;
end;

procedure T8085.ExecuteLine;
var
 opcode,tmp1,tmp2:integer;
begin
 if fstart=false then exit;
 opcode:=fmemory[fregpc];
 case opcode of
  $00:
    begin
     fregpc:=fregpc+1;
    end;
  $01:
    begin
     fregc:=fmemory[fregpc+1];
     fregb:=fmemory[fregpc+2];
     fregpc:=fregpc+3;
    end;
  $02:
    begin
     tmp1:=((fregb shl 8) or fregc) and $FFFF;
     fmemory[tmp1]:=frega;
     fregpc:=fregpc+1;
    end;
  $03:
    begin
     tmp1:=((fregb shl 8) or fregc) and $FFFF;
     tmp1:=tmp1+1;
     fregb:=(tmp1 shr 8) and $FF;
     fregc:=tmp1 and $FF;
     fregpc:=fregpc+1;
    end;
  $04:
    begin
     fregb:=IInr(fregb,1);
     fregpc:=fregpc+1;
    end;
  $05:
    begin
     fregb:=IInr(fregb,-1);
     fregpc:=fregpc+1;
    end;
  $06:
    begin
     fregb:=fmemory[fregpc+1];
     fregpc:=fregpc+2;
    end;
  $07:
   begin
    tmp1:=frega and $80;
    frega:=frega shl 1 or tmp1;
    fregpsw:=fregpsw or tmp1;
    fregpc:=fregpc+1;
   end;
  $09:
   begin
    tmp1:=((fregh shl 8) or fregl) and $FFFF;
    tmp2:=((fregb shl 8) or fregc) and $FFFF;
    tmp1:=(tmp1+tmp2);
    if tmp1>$FFFF then fregpsw:=fregpsw or 1 else fregpsw :=fregpsw and not $1; 
    fregh:=(tmp1 shr 8) and $FF;
    fregl:=tmp1 and $FF;
    fregpc:=fregpc+1;
   end;
  $0A:
   begin
    tmp1:=((fregb shl 8) or fregc) and $FFFF;
    frega:=fmemory[tmp1];
    fregpc:=fregpc+1;
   end;
  $0B:
   begin
    tmp1:=(((fregb shl 8) or fregc) and $FFFF)-1;
    fregb:=(tmp1 shr 8) and $FF;
    fregc:=tmp1 and $FF;
    fregpc:=fregpc+1;
   end;
  $0C:
    begin
     fregc:=IInr(fregc,1);
     fregpc:=fregpc+1;
    end;
  $0D:
    begin
     fregc:=IInr(fregc,-1);
     fregpc:=fregpc+1;
    end;
  $0E:
   begin
    fregc:=fmemory[fregpc+1];
    fregpc:=fregpc+2;
   end;
  $0F:
   begin
    tmp1:=frega and $1;
    fregpsw:=fregpsw or tmp1;
    frega:=(frega shr 1) or (tmp1 shl 7);
    fregpc:=fregpc+1;
   end;
  $11:
    begin
     frege:=fmemory[fregpc+1];
     fregd:=fmemory[fregpc+2];
     fregpc:=fregpc+3;
    end;
  $12:
    begin
     tmp1:=((fregd shl 8) or frege) and $FFFF;
     fmemory[tmp1]:=frega;
     fregpc:=fregpc+1;
    end;
  $13:
    begin
     tmp1:=((fregd shl 8) or frege) and $FFFF;
     tmp1:=tmp1+1;
     fregd:=(tmp1 shr 8) and $FF;
     frege:=tmp1 and $FF;
     fregpc:=fregpc+1;
    end;
  $14:
    begin
     fregd:=IInr(fregd,1);
     fregpc:=fregpc+1;
    end;
  $15:
    begin
     fregd:=IInr(fregd,-1);
     fregpc:=fregpc+1;
    end;
  $16:
   begin
    fregd:=fmemory[fregpc+1];
    fregpc:=fregpc+2;
   end;
  $17:
   begin
    tmp1:=frega and $80;
    frega:=frega shl 1 or (fregpsw and $1);
    if tmp1=$80 then fregpsw:=fregpsw or $1
    else fregpsw:=fregpsw and not $1;
    fregpc:=fregpc+1;
   end;
  $19:
   begin
    tmp1:=((fregh shl 8) or fregl) and $FFFF;
    tmp2:=((fregd shl 8) or frege) and $FFFF;
    tmp1:=(tmp1+tmp2);
    if tmp1>$FFFF then fregpsw:=fregpsw or 1 else fregpsw :=fregpsw and not $1;    
    fregh:=(tmp1 shr 8) and $FF;
    fregl:=tmp1 and $FF;
    fregpc:=fregpc+1;
   end;
  $1A:
   begin
    tmp1:=((fregd shl 8) or frege) and $FFFF;
    frega:=fmemory[tmp1];
    fregpc:=fregpc+1;
   end;
  $1B:
   begin
    tmp1:=(((fregd shl 8) or frege) and $FFFF)-1;
    fregd:=(tmp1 shr 8) and $FF;
    frege:=tmp1 and $FF;
    fregpc:=fregpc+1;
   end;
  $1C:
    begin
     frege:=IInr(frege,1);
     fregpc:=fregpc+1;
    end;
  $1D:
    begin
     frege:=IInr(frege,-1);
     fregpc:=fregpc+1;
    end;
  $1E:
   begin
    frege:=fmemory[fregpc+1];
    fregpc:=fregpc+2;
   end;
  $1F:
   begin
    tmp1:=frega and $1;tmp2:=fregpsw and $1;
    frega:=(frega shr 1) or (tmp2 shl 7);
    if tmp1=$1 then fregpsw:=fregpsw or $1
    else fregpsw:=fregpsw and not $1;
    fregpc:=fregpc+1;
   end;
  $21:
   begin
    fregl:=fmemory[fregpc+1];
    fregh:=fmemory[fregpc+2];
    fregpc:=fregpc+3;
   end;
  $22:
   begin
    tmp1:=((fmemory[fregpc+2] shl 8) or fmemory[fregpc+1]) and $FFFF;
    fmemory[tmp1]:=fregl;
    fmemory[tmp1+1]:=fregh;
    fregpc:=fregpc+3;
   end;
  $23:
    begin
     tmp1:=((fregh shl 8) or fregl) and $FFFF;
     tmp1:=tmp1+1;
     fregh:=(tmp1 shr 8) and $FF;
     fregl:=tmp1 and $FF;
     fregpc:=fregpc+1;
    end;
  $24:
    begin
     fregh:=IInr(fregh,1);
     fregpc:=fregpc+1;
    end;
  $25:
    begin
     fregh:=IInr(fregh,-1);
     fregpc:=fregpc+1;
    end;
  $26:
   begin
    fregh:=fmemory[fregpc+1];
    fregpc:=fregpc+2;
   end;
  $27:
   begin
    tmp1:=frega and $F;tmp2:=0;
    if (tmp1>9) or (fregpsw and $10>0) then
     begin
      tmp1:=tmp1+6;
      if tmp1>$F then tmp2:=1;
      tmp1:=tmp1 and $F;
      frega:=(frega and $F0) or tmp1;
      if tmp2=1 then fregpsw:=fregpsw or $10;
     end;
    tmp1:=(frega shr 4) and $F;
    tmp1:=(tmp1+tmp2);
    if (tmp1>9) or (fregpsw and $1>0) then
     begin
      tmp1:=tmp1+6;
      if(tmp1>$F) then fregpsw:=fregpsw or $1;
      tmp1:=tmp1 and $F;
     end;
    frega:=(frega and $0F) or (tmp1 shl 4);
    if(frega=0) then fregpsw:=fregpsw or $40 else fregpsw:=fregpsw and not $40;
    if(frega and $80)>0 then fregpsw:=fregpsw or $80 else fregpsw:=fregpsw and not $80;
    if GetParity(frega)=1 then fregpsw:=fregpsw or $4 else fregpsw:=fregpsw and not $4;
    fregpc:=fregpc+1;
   end;
  $29:
   begin
    tmp1:=((fregh shl 8) or fregl) and $FFFF;
    tmp2:=((fregh shl 8) or fregl) and $FFFF;
    tmp1:=(tmp1+tmp2);
    if tmp1>$FFFF then fregpsw:=fregpsw or 1 else fregpsw :=fregpsw and not $1;
    fregh:=(tmp1 shr 8) and $FF;
    fregl:=tmp1 and $FF;
    fregpc:=fregpc+1;
   end;
  $2A:
   begin
    tmp1:=((fmemory[fregpc+2] shl 8) or fmemory[fregpc+1]) and $FFFF;
    fregl:=fmemory[tmp1];
    fregh:=fmemory[tmp1+1];
    fregpc:=fregpc+2;
   end;
  $2B:
   begin
   tmp1:=(((fregh shl 8) or fregl) and $FFFF)-1;
   fregh:=(tmp1 shr 8) and $FF;
   fregl:=tmp1 and $FF;
   fregpc:=fregpc+1;
   end;
  $2C:
    begin
     fregl:=IInr(fregl,1);
     fregpc:=fregpc+1;
    end;
  $2D:
    begin
     fregl:=IInr(fregl,-1);
     fregpc:=fregpc+1;
    end;
  $2E:
   begin
    fregl:=fmemory[fregpc+1];
    fregpc:=fregpc+2;
   end;
  $2F:
   begin
    frega:=not frega;
    fregpc:=fregpc+1;
   end;
  $31:
   begin
    fregsp:=fmemory[fregpc+2] shl 8 or fmemory[fregpc+1];
    fregpc:=fregpc+3;
   end;
  $32:
   begin
    tmp1:=(fmemory[fregpc+2] shl 8 or fmemory[fregpc+1]) and $FFFF;
    fmemory[tmp1]:=frega;
    fregpc:=fregpc+3;
   end;
  $33:
    begin
     fregsp:=fregsp+1;
     fregpc:=fregpc+1;
    end;
  $34:
    begin
     tmp1:=fmemory[fregh shl 8 or fregl];
     tmp1:=IInr(tmp1,1);
     fmemory[fregh shl 8 or fregl]:=tmp1;
     fregpc:=fregpc+1;
    end;
  $35:
    begin
     tmp1:=fmemory[fregh shl 8 or fregl];
     tmp1:=IInr(tmp1,-1);
     fmemory[fregh shl 8 or fregl]:=tmp1;
     fregpc:=fregpc+1;
    end;
  $36:
   begin
    tmp1:=fregh shl 8 or fregl;
    fmemory[tmp1]:=fmemory[fregpc+1];
    fregpc:=fregpc+2;
   end;
  $37:
   begin
    fregpsw:=fregpsw or $1;
    fregpc:=fregpc+1;
   end;
  $39:
   begin
    tmp1:=((fregh shl 8) or fregl) and $FFFF;
    tmp1:=(tmp1+fregsp);
    fregh:=(tmp1 shr 8) and $FF;
    fregl:=tmp1 and $FF;
    fregpc:=fregpc+1;
   end;
  $3A:
   begin
    tmp1:=((fmemory[fregpc+2] shl 8) or fmemory[fregpc+1]) and $FFFF;
    frega:=fmemory[tmp1];
    fregpc:=fregpc+3;
   end;
  $3B:
   begin
   fregsp:=fregsp-1;
   fregpc:=fregpc+1;
   end;
  $3C:
    begin
     frega:=IInr(frega,1);
     fregpc:=fregpc+1;
    end;
  $3D:
   begin
    frega:=IInr(frega,-1);
    fregpc:=fregpc+1;
   end;
  $3E:
   begin
    frega:=fmemory[fregpc+1];
    fregpc:=fregpc+2;
   end;
  $3F:
   begin
    if (fregpsw and $1)=0 then fregpsw:=fregpsw or 1
    else fregpsw:=fregpsw and not $1;
    fregpc:=fregpc+1;
   end;
  $40:
   begin
    fregpc:=fregpc+1;
   end;
  $41:
   begin
    fregb:=fregc;
    fregpc:=fregpc+1;
   end;
  $42:
   begin
    fregb:=fregd;
    fregpc:=fregpc+1;
   end;
  $43:
   begin
    fregb:=frege;
    fregpc:=fregpc+1;
   end;
  $44:
   begin
    fregb:=fregh;
    fregpc:=fregpc+1;
   end;
  $45:
   begin
    frega:=fregl;
    fregpc:=fregpc+1;
   end;
  $46:
   begin
    tmp1:=fregh shl 8 or fregl;
    fregb:=fmemory[tmp1];
    fregpc:=fregpc+1;
   end;
  $47:
   begin
    fregb:=frega;
    fregpc:=fregpc+1;
   end;
  $48:
   begin
    fregc:=fregb;
    fregpc:=fregpc+1;
   end;
  $49:
   begin
    fregpc:=fregpc+1;
   end;
  $4A:
   begin
    fregc:=fregd;
    fregpc:=fregpc+1;
   end;
  $4B:
   begin
    fregc:=frege;
    fregpc:=fregpc+1;
   end;
  $4C:
   begin
    fregc:=fregh;
    fregpc:=fregpc+1;
   end;
  $4D:
   begin
    fregc:=fregl;
    fregpc:=fregpc+1;
   end;
  $4E:
   begin
    tmp1:=fregh shl 8 or fregl;
    fregc:=fmemory[tmp1];
    fregpc:=fregpc+1;
   end;
  $4F:
   begin
    fregc:=frega;
    fregpc:=fregpc+1;
   end;
  $50:
   begin
    fregd:=fregb;
    fregpc:=fregpc+1;
   end;
  $51:
   begin
    fregd:=fregc;
    fregpc:=fregpc+1;
   end;
  $52:
   begin
    fregpc:=fregpc+1;
   end;
  $53:
   begin
    fregd:=frege;
    fregpc:=fregpc+1;
   end;
  $54:
   begin
    fregd:=fregh;
    fregpc:=fregpc+1;
   end;
  $55:
   begin
    fregd:=fregl;
    fregpc:=fregpc+1;
   end;
  $56:
   begin
    tmp1:=fregh shl 8 or fregl;
    fregd:=fmemory[tmp1];
    fregpc:=fregpc+1;
   end;
  $57:
   begin
    fregd:=frega;
    fregpc:=fregpc+1;
   end;
  $58:
   begin
    frege:=fregb;
    fregpc:=fregpc+1;
   end;
  $59:
   begin
    frege:=fregc;
    fregpc:=fregpc+1;
   end;
  $5A:
   begin
    frege:=fregd;
    fregpc:=fregpc+1;
   end;
  $5B:
   begin
    fregpc:=fregpc+1;
   end;
  $5C:
   begin
    frege:=fregh;
    fregpc:=fregpc+1;
   end;
  $5D:
   begin
    frege:=fregl;
    fregpc:=fregpc+1;
   end;
  $5E:
   begin
    tmp1:=fregh shl 8 or fregl;
    frege:=fmemory[tmp1];
    fregpc:=fregpc+1;
   end;
  $5F:
   begin
    frege:=frega;
    fregpc:=fregpc+1;
   end;
  $60:
   begin
    fregh:=fregb;
    fregpc:=fregpc+1;
   end;
  $61:
   begin
    fregh:=fregc;
    fregpc:=fregpc+1;
   end;
  $62:
   begin
    fregh:=fregd;
    fregpc:=fregpc+1;
   end;
  $63:
   begin
    fregh:=frege;
    fregpc:=fregpc+1;
   end;
  $64:
   begin
    fregpc:=fregpc+1;
   end;
  $65:
   begin
    fregh:=fregl;
    fregpc:=fregpc+1;
   end;
  $66:
   begin
    tmp1:=fregh shl 8 or fregl;
    fregh:=fmemory[tmp1];
    fregpc:=fregpc+1;
   end;
  $67:
   begin
    fregh:=frega;
    fregpc:=fregpc+1;
   end;
  $68:
   begin
    fregl:=fregb;
    fregpc:=fregpc+1;
   end;
  $69:
   begin
    fregl:=fregc;
    fregpc:=fregpc+1;
   end;
  $6A:
   begin
    fregl:=fregd;
    fregpc:=fregpc+1;
   end;
  $6B:
   begin
    fregl:=frege;
    fregpc:=fregpc+1;
   end;
  $6C:
   begin
    fregl:=fregh;
    fregpc:=fregpc+1;
   end;
  $6D:
   begin
    fregpc:=fregpc+1;
   end;
  $6E:
   begin
    tmp1:=fregh shl 8 or fregl;
    fregl:=fmemory[tmp1];
    fregpc:=fregpc+1;
   end;
  $6F:
   begin
    fregl:=frega;
    fregpc:=fregpc+1;
   end;
  $70:
   begin
    tmp1:=fregh shl 8 or fregl;
    fmemory[tmp1]:=fregb;
    fregpc:=fregpc+1;
   end;
  $71:
   begin
    tmp1:=fregh shl 8 or fregl;
    fmemory[tmp1]:=fregc;
    fregpc:=fregpc+1;
   end;
  $72:
   begin
    tmp1:=fregh shl 8 or fregl;
    fmemory[tmp1]:=fregd;
    fregpc:=fregpc+1;
   end;
  $73:
   begin
    tmp1:=fregh shl 8 or fregl;
    fmemory[tmp1]:=frege;
    fregpc:=fregpc+1;
   end;
  $74:
   begin
    tmp1:=fregh shl 8 or fregl;
    fmemory[tmp1]:=fregh;
    fregpc:=fregpc+1;
   end;
  $75:
   begin
    tmp1:=fregh shl 8 or fregl;
    fmemory[tmp1]:=fregl;
    fregpc:=fregpc+1;
   end;
  $76:
   begin
    fstart:=false;fret:=false;
    fcall:=false;frec:=0;
   end;
  $77:
   begin
    tmp1:=fregh shl 8 or fregl;
    fmemory[tmp1]:=frega;
    fregpc:=fregpc+1;
   end;
  $78:
   begin
    frega:=fregb;
    fregpc:=fregpc+1;
   end;
  $79:
   begin
    frega:=fregc;
    fregpc:=fregpc+1;
   end;
  $7A:
   begin
    frega:=fregd;
    fregpc:=fregpc+1;
   end;
  $7B:
   begin
    frega:=frege;
    fregpc:=fregpc+1;
   end;
  $7C:
   begin
    frega:=fregh;
    fregpc:=fregpc+1;
   end;
  $7D:
   begin
    frega:=fregl;
    fregpc:=fregpc+1;
   end;
  $7E:
   begin
    tmp1:=fregh shl 8 or fregl;
    frega:=fmemory[tmp1];
    fregpc:=fregpc+1;
   end;
  $7F:
   begin
    fregpc:=fregpc+1;
   end;
  $80:
   begin
    IAdd(fregb,0);
    fregpc:=fregpc+1;
   end;
  $81:
   begin
    IAdd(fregc,0);
    fregpc:=fregpc+1;
   end;
  $82:
   begin
    IAdd(fregd,0);
    fregpc:=fregpc+1;
   end;
  $83:
   begin
    IAdd(frege,0);
    fregpc:=fregpc+1;
   end;
  $84:
   begin
    IAdd(fregh,0);
    fregpc:=fregpc+1;
   end;
  $85:
   begin
    IAdd(fregl,0);
    fregpc:=fregpc+1;
   end;
  $86:
   begin
    tmp1:=fregh shl 8 or fregl;
    IAdd(fmemory[tmp1],0);
    fregpc:=fregpc+1;
   end;
  $87:
   begin
    IAdd(frega,0);
    fregpc:=fregpc+1;
   end;
  $88:
   begin
    IAdd(fregb,fregpsw and $1);
    fregpc:=fregpc+1;
   end;
  $89:
   begin
    IAdd(fregc,fregpsw and $1);
    fregpc:=fregpc+1;
   end;
  $8A:
   begin
    IAdd(fregd,fregpsw and $1);
    fregpc:=fregpc+1;
   end;
  $8B:
   begin
    IAdd(frege,fregpsw and $1);
    fregpc:=fregpc+1;
   end;
  $8C:
   begin
    IAdd(fregh,fregpsw and $1);
    fregpc:=fregpc+1;
   end;
  $8D:
   begin
    IAdd(fregl,fregpsw and $1);
    fregpc:=fregpc+1;
   end;
  $8E:
   begin
    tmp1:=fregh shl 8 or fregl;
    IAdd(fmemory[tmp1],fregpsw and $1);
    fregpc:=fregpc+1;
   end;
  $8F:
   begin
    IAdd(frega,fregpsw and $1);
    fregpc:=fregpc+1;
   end;
  $90:
   begin
    Isub(fregb,0);
    fregpc:=fregpc+1;
   end;
  $91:
   begin
    Isub(fregc,0);
    fregpc:=fregpc+1;
   end;
  $92:
   begin
    Isub(fregd,0);
    fregpc:=fregpc+1;
   end;
  $93:
   begin
    Isub(frege,0);
    fregpc:=fregpc+1;
   end;
  $94:
   begin
    Isub(fregh,0);
    fregpc:=fregpc+1;
   end;
  $95:
   begin
    Isub(fregl,0);
    fregpc:=fregpc+1;
   end;
  $96:
   begin
    tmp1:=fregh shl 8 or fregl;
    Isub(fmemory[tmp1],0);
    fregpc:=fregpc+1;
   end;
  $97:
   begin
    Isub(frega,0);
    fregpc:=fregpc+1;
   end;
  $98:
   begin
    Isub(fregb,fregpsw and $1);
    fregpc:=fregpc+1;
   end;
  $99:
   begin
    Isub(fregc,fregpsw and $1);
    fregpc:=fregpc+1;
   end;
  $9A:
   begin
    Isub(fregd,fregpsw and $1);
    fregpc:=fregpc+1;
   end;
  $9B:
   begin
    Isub(frege,fregpsw and $1);
    fregpc:=fregpc+1;
   end;
  $9C:
   begin
    Isub(fregh,fregpsw and $1);
    fregpc:=fregpc+1;
   end;
  $9D:
   begin
    Isub(fregl,fregpsw and $1);
    fregpc:=fregpc+1;
   end;
  $9E:
   begin
    tmp1:=fregh shl 8 or fregl;
    Isub(fmemory[tmp1],fregpsw and $1);
    fregpc:=fregpc+1;
   end;
  $9F:
   begin
    Isub(frega,fregpsw and $1);
    fregpc:=fregpc+1;
   end;
  $A0:
   begin
    IAnd(fregb);
    fregpc:=fregpc+1;
   end;
  $A1:
   begin
    IAnd(fregc);
    fregpc:=fregpc+1;
   end;
  $A2:
   begin
    IAnd(fregd);
    fregpc:=fregpc+1;
   end;
  $A3:
   begin
    IAnd(frege);
    fregpc:=fregpc+1;
   end;
  $A4:
   begin
    IAnd(fregh);
    fregpc:=fregpc+1;
   end;
  $A5:
   begin
    IAnd(fregl);
    fregpc:=fregpc+1;
   end;
  $A6:
   begin
    tmp1:=fregh shl 8 or fregl;
    IAnd(fmemory[tmp1]);
    fregpc:=fregpc+1;
   end;
  $A7:
   begin
    IAnd(frega);
    fregpc:=fregpc+1;
   end;
  $A8:
   begin
    IXor(fregb);
    fregpc:=fregpc+1;
   end;
  $A9:
   begin
    IXor(fregc);
    fregpc:=fregpc+1;
   end;
  $AA:
   begin
    IXor(fregd);
    fregpc:=fregpc+1;
   end;
  $AB:
   begin
    IXor(frege);
    fregpc:=fregpc+1;
   end;
  $AC:
   begin
    IXor(fregh);
    fregpc:=fregpc+1;
   end;
  $AD:
   begin
    IXor(fregl);
    fregpc:=fregpc+1;
   end;
  $AE:
   begin
    tmp1:=fregh shl 8 or fregl;
    IXor(fmemory[tmp1]);
    fregpc:=fregpc+1;
   end;
  $AF:
   begin
    IXor(frega);
    fregpc:=fregpc+1;
   end;
  $B0:
   begin
    Ior(fregb);
    fregpc:=fregpc+1;
   end;
  $B1:
   begin
    Ior(fregc);
    fregpc:=fregpc+1;
   end;
  $B2:
   begin
    Ior(fregd);
    fregpc:=fregpc+1;
   end;
  $B3:
   begin
    Ior(frege);
    fregpc:=fregpc+1;
   end;
  $B4:
   begin
    Ior(fregh);
    fregpc:=fregpc+1;
   end;
  $B5:
   begin
    Ior(fregl);
    fregpc:=fregpc+1;
   end;
  $B6:
   begin
    tmp1:=fregh shl 8 or fregl;
    Ior(fmemory[tmp1]);
    fregpc:=fregpc+1;
   end;
  $B7:
   begin
    Ior(frega);
    fregpc:=fregpc+1;
   end;
  $B8:
   begin
    Icmp(fregb);
    fregpc:=fregpc+1;
   end;
  $B9:
   begin
    Icmp(fregc);
    fregpc:=fregpc+1;
   end;
  $BA:
   begin
    Icmp(fregd);
    fregpc:=fregpc+1;
   end;
  $BB:
   begin
    Icmp(frege);
    fregpc:=fregpc+1;
   end;
  $BC:
   begin
    Icmp(fregh);
    fregpc:=fregpc+1;
   end;
  $BD:
   begin
    Icmp(fregl);
    fregpc:=fregpc+1;
   end;
  $BE:
   begin
    tmp1:=(fregh shl 8) or fregl;
    Icmp(fmemory[tmp1]);
    fregpc:=fregpc+1;
   end;
  $BF:
   begin
    Icmp(frega);
    fregpc:=fregpc+1;
   end;
  $C0:
   begin
    if(fregpsw and $40)=0 then Iret else fregpc:=fregpc+1;
   end;
  $C1:
   begin
    fregc:=fmemory[fregsp];
    fregb:=fmemory[fregsp+1];
    fregsp:=fregsp+2;
    fregpc:=fregpc+1;
   end;
  $C2:
   begin
    if(fregpsw and $40)=0 then
     fregpc:=(fmemory[fregpc+2] shl 8 or fmemory[fregpc+1]) and $FFFF
    else
     fregpc:=fregpc+3;
   end;
  $C3:
   begin
    fregpc:=(fmemory[fregpc+2] shl 8 or fmemory[fregpc+1]) and $FFFF;
   end;
  $C4:
   begin
    if (fregpsw and $40)=0 then
     begin
      tmp1:=(fmemory[fregpc+2] shl 8) or fmemory[fregpc+1];
      ICall(tmp1);
     end
    else
     fregpc:=fregpc+1;
   end;
  $C5:
   begin
    fmemory[fregsp-1]:=fregb;
    fmemory[fregsp-2]:=fregc;
    fregsp:=fregsp-2;
    fregpc:=fregpc+1;
   end;
  $C6:
   begin
    IAdd(fmemory[fregpc+1],0);
    fregpc:=fregpc+2;
   end;
  $C7:
   begin
    fregpc:=fregpc+1;
    fmemory[fregsp-1]:=fregpc shr 8 and $FF;
    fmemory[fregsp-2]:=fregpc and $FF;
    fregsp:=fregsp-2;
    fregpc:=$0000;
   end;
  $C8:
   begin
    if(fregpsw and $40)=1 then Iret else fregpc:=fregpc+1;
   end;
  $C9:
   begin
    Iret;
   end;
  $CA:
   begin
    if(fregpsw and $40)>0 then
     fregpc:=(fmemory[fregpc+2] shl 8 or fmemory[fregpc+1]) and $FFFF
    else
     fregpc:=fregpc+3;
   end;
  $CC:
   begin
    if (fregpsw and $40)=1 then
     begin
      tmp1:=(fmemory[fregpc+2] shl 8) or fmemory[fregpc+1];
      ICall(tmp1);
     end
    else
     fregpc:=fregpc+1;
   end;
  $CD:
   begin
    tmp1:=(fmemory[fregpc+2] shl 8) or fmemory[fregpc+1];
    ICall(tmp1);
   end;
  $CE:
   begin
    IAdd(fmemory[fregpc+1],fregpsw and $1);
    fregpc:=fregpc+2;
   end;
  $CF:
   begin
    fregpc:=fregpc+1;
    fmemory[fregsp-1]:=fregpc shr 8 and $FF;
    fmemory[fregsp-2]:=fregpc and $FF;
    fregsp:=fregsp-2;
    fregpc:=$0008;
   end;
  $D0:
   begin
    if(fregpsw and $1)=0 then Iret else fregpc:=fregpc+1;
   end;
  $D1:
   begin
    frege:=fmemory[fregsp];
    fregd:=fmemory[fregsp+1];
    fregsp:=fregsp+2;
    fregpc:=fregpc+1;
   end;
  $D2:
   begin
    if(fregpsw and $1)=0 then
     fregpc:=(fmemory[fregpc+2] shl 8 or fmemory[fregpc+1]) and $FFFF
    else
     fregpc:=fregpc+3;
   end;
  $D4:
   begin
    if (fregpsw and $1)=0 then
     begin
      tmp1:=(fmemory[fregpc+2] shl 8) or fmemory[fregpc+1];
      ICall(tmp1);
     end
    else
     fregpc:=fregpc+1;
   end;
  $D5:
   begin
    fmemory[fregsp-1]:=fregd;
    fmemory[fregsp-2]:=frege;
    fregsp:=fregsp-2;
    fregpc:=fregpc+1;
   end;
  $D6:
   begin
    ISub(fmemory[fregpc+1],0);
    fregpc:=fregpc+2;
   end;
  $D7:
   begin
    fregpc:=fregpc+1;
    fmemory[fregsp-1]:=fregpc shr 8 and $FF;
    fmemory[fregsp-2]:=fregpc and $FF;
    fregsp:=fregsp-2;
    fregpc:=$0010;
   end;
  $D8:
   begin
    if(fregpsw and $1)>0 then Iret else fregpc:=fregpc+1;
   end;
  $DA:
   begin
    if(fregpsw and $1)>0 then
     fregpc:=(fmemory[fregpc+2] shl 8 or fmemory[fregpc+1]) and $FFFF
    else
     fregpc:=fregpc+3;
   end;
  $DC:
   begin
    if (fregpsw and $1)>0 then
     begin
      tmp1:=(fmemory[fregpc+2] shl 8) or fmemory[fregpc+1];
      ICall(tmp1);
     end
    else
     fregpc:=fregpc+1;
   end;
  $DE:
   begin
    ISub(fmemory[fregpc+1],fregpsw and $1);
    fregpc:=fregpc+2;
   end;
  $DF:
   begin
    fregpc:=fregpc+1;
    fmemory[fregsp-1]:=fregpc shr 8 and $FF;
    fmemory[fregsp-2]:=fregpc and $FF;
    fregsp:=fregsp-2;
    fregpc:=$0018;
   end;
  $E0:
   begin
    if(fregpsw and $4)=0 then Iret else fregpc:=fregpc+1;
   end;
  $E1:
   begin
    fregl:=fmemory[fregsp];
    fregh:=fmemory[fregsp+1];
    fregsp:=fregsp+2;
    fregpc:=fregpc+1;
   end;
  $E2:
   begin
    if(fregpsw and $4)=0 then
     fregpc:=(fmemory[fregpc+2] shl 8 or fmemory[fregpc+1]) and $FFFF
    else
     fregpc:=fregpc+3;
   end;
  $E3:
   begin
    tmp1:=(fregsp shr 8) and $FF;
    tmp2:=fregsp and $FF;
    fregsp:=fregh shl 8 or fregl;
    fregh:=tmp1;fregl:=tmp2; 
    fregpc:=fregpc+1;
   end;
  $E4:
   begin
    if (fregpsw and $4)=0 then
     begin
      tmp1:=(fmemory[fregpc+2] shl 8) or fmemory[fregpc+1];
      ICall(tmp1);
     end
    else
     fregpc:=fregpc+1;
   end;
  $E5:
   begin
    fmemory[fregsp-1]:=fregh;
    fmemory[fregsp-2]:=fregl;
    fregsp:=fregsp-2;
    fregpc:=fregpc+1;
   end;
  $E6:
   begin
    IAnd(fmemory[fregpc+1]);
    fregpc:=fregpc+2;
   end;
  $E7:
   begin
    fregpc:=fregpc+1;
    fmemory[fregsp-1]:=fregpc shr 8 and $FF;
    fmemory[fregsp-2]:=fregpc and $FF;
    fregsp:=fregsp-2;
    fregpc:=$0020;
   end;
  $E8:
   begin
    if(fregpsw and $4)>0 then Iret else fregpc:=fregpc+1;
   end;
  $E9:
   begin
    fregpc:=(fregh shl 8) or fregl;
   end;
  $EA:
   begin
    if(fregpsw and $4)>0 then
     fregpc:=(fmemory[fregpc+2] shl 8 or fmemory[fregpc+1]) and $FFFF
    else
     fregpc:=fregpc+3;
   end;
  $EB:
   begin
    tmp1:=fregd;fregd:=fregh;fregh:=tmp1;
    tmp1:=frege;frege:=fregl;fregl:=tmp1;
    fregpc:=fregpc+1;
   end;
  $EC:
   begin
    if (fregpsw and $4)=1 then
     begin
      tmp1:=(fmemory[fregpc+2] shl 8) or fmemory[fregpc+1];
      ICall(tmp1);
     end
    else
     fregpc:=fregpc+1;
   end;
  $EE:
   begin
    IXor(fmemory[fregpc+1]);
    fregpc:=fregpc+2;
   end;
  $EF:
   begin
    fregpc:=fregpc+1;
    fmemory[fregsp-1]:=fregpc shr 8 and $FF;
    fmemory[fregsp-2]:=fregpc and $FF;
    fregsp:=fregsp-2;
    fregpc:=$0028;
   end;
  $F0:
   begin
    if(fregpsw and $80)=0 then Iret else fregpc:=fregpc+1;
   end;
  $F1:
   begin
    fregpsw:=fmemory[fregsp];
    frega:=fmemory[fregsp+1];
    fregsp:=fregsp+2;
    fregpc:=fregpc+1;
   end;
  $F2:
   begin
    if(fregpsw and $80)=0 then
     fregpc:=(fmemory[fregpc+2] shl 8 or fmemory[fregpc+1]) and $FFFF
    else
     fregpc:=fregpc+3;
   end;
  $F4:
   begin
    if (fregpsw and $80)=0 then
     begin
      tmp1:=(fmemory[fregpc+2] shl 8) or fmemory[fregpc+1];
      ICall(tmp1);
     end
    else
     fregpc:=fregpc+1;
   end;
  $F5:
   begin
    fmemory[fregsp-1]:=frega;
    fmemory[fregsp-2]:=fregpsw;
    fregsp:=fregsp-2;
    fregpc:=fregpc+1;
   end;
  $F6:
   begin
    Ior(fmemory[fregpc+1]);
    fregpc:=fregpc+2;
   end;
  $F7:
   begin
    fregpc:=fregpc+1;
    fmemory[fregsp-1]:=fregpc shr 8 and $FF;
    fmemory[fregsp-2]:=fregpc and $FF;
    fregsp:=fregsp-2;
    fregpc:=$0030;
   end;
  $F8:
   begin
    if(fregpsw and $80)>0 then Iret else fregpc:=fregpc+1;
   end;
  $F9:
   begin
    tmp1:=fregh shl 8 or fregl;
    fregsp:=tmp1;
    fregpc:=fregpc+1;
   end;
  $FA:
   begin
    if(fregpsw and $80)>0 then
     fregpc:=(fmemory[fregpc+2] shl 8 or fmemory[fregpc+1]) and $FFFF
    else
     fregpc:=fregpc+3;
   end;
  $FC:
   begin
    if (fregpsw and $80)>0 then
     begin
      tmp1:=(fmemory[fregpc+2] shl 8) or fmemory[fregpc+1];
      ICall(tmp1);
     end
    else
     fregpc:=fregpc+3;
   end;
  $FE:
   begin
    Icmp(fmemory[fregpc+1]);
    fregpc:=fregpc+2;
   end;
  $FF:
   begin
    fregpc:=fregpc+1;
    fmemory[fregsp-1]:=fregpc shr 8 and $FF;
    fmemory[fregsp-2]:=fregpc and $FF;
    fregsp:=fregsp-2;
    fregpc:=$0038;
   end;
  $08,$10,$18,$28,$38,$CB,$DD,$ED,$FD:
   begin
    fregpc:=fregpc+1;
   end;
 end;
end;


function HexToDec(c:char):integer;
begin
 result:=-1;
 if(c>='0') and (c<='9') then result:=Hex1[c]
 else
  begin
   c:=char(byte(c) and $DF);
   if(c>='A') and (c<='F') then result:=Hex2[c];
  end;
end;

function HexToDec(s:pchar):integer;
var
 i:integer;
begin
 result:=-1;if(s=nil) then exit;
 result:=0;
 for i:=0 to strlen(s)-1 do result:=result*16+HexToDec(s[i]);
end;

procedure DecToHex(data:word;s:pchar);
begin
 if s=nil then exit;
 s[3]:=Dec[data and $F];
 s[2]:=Dec[data shr 4 and $F];
 s[1]:=Dec[data shr 8 and $F];
 s[0]:=Dec[data shr 12 and $F];
end;

function GetParity(a:byte):integer;
var
 i:integer;
begin
 result:=0;
 for i:=0 to 7 do result:=(result + (a shr i) and $1) mod 2;
 result:=(result+1) mod 2;
end;
end.
