unit ULex;

interface
uses
 classes,sysutils,windows;

type
 TAlphabetSet=array[0..15] of byte;
 PState=^TState;
 TState=record
  eat:boolean;
  ch:TAlphabetSet;
  accept:smallint;
  cinfo,lh:integer;
  next1,next2:PState;
 end;

 PMachine=^TMachine;
 TMachine=record
  start:PState;
  final:PState;
 end;

 PTokenNode=^TTokenNode;
 TTokenNode=record
  tag:char;
  data:PMachine;
  next:PTokenNode;
 end;

 TToken= record
  start:integer;
  stop :integer;
  code :integer;
 end;
 PToken=^TToken;

 TStack=array[0..2047] of integer;
 PStack=^TStack;

 TSpace=array[0..2047] of TState;
 PSpace=^TSpace;

type
 TMList=class
  private
   ffront,frear:PTokenNode;
  public
   function InsertMachine(adata:PMachine):boolean;
   function InsertOperator(aop:char):boolean;
   function RemoveMachine:PMachine;
   function RemoveOperator:char;
   function IsEmpty:boolean;
   function IsMachine:boolean;
   procedure Clear;
   destructor Destroy;override;
  end;

 TLex=class
  private
   fspace:PSpace;
   fstack:PStack;
   flaccept,faccept:PState;
   ftos,fcount:integer;
   forigin:longword;
   flist:TMList;
   fbeginset:TBits;
   fclosureset:TBits;
   fmoveset:TBits;
   fmachine:PMachine;
  protected
   function CreateMachine(ainfo:integer;aset:TAlphabetSet):PMachine;
   function OrMachine(mac1,mac2:PMachine):PMachine;
   function AndMachine(mac1,mac2:PMachine):PMachine;
   function KCloseMachine(amac:PMachine):PMachine;
   function PCloseMachine(amac:PMachine):PMachine;
   function QCloseMachine(amac:PMachine):PMachine;
   function GetState:PState;
   function GetClosureSet(aset:TBits):boolean;
   procedure GetMoveSet(aset:TBits;ach:char);
   function ConvertToken(atoken:string):boolean;
   function ConstructMachine:PMachine;
  public
   constructor Create;
   function AddToken(atoken:string;index:integer;lh:integer=0;eat:boolean=false):boolean;
   procedure ClearTokens;
   function ListTokens(aline:string;var tknlist:array of TToken;tkncnt:integer):integer;
   destructor Destroy;override;
  end;

function IsSetEmpty(aset:TAlphabetSet):boolean;
function IsAscii(achar:char):boolean;
function IsChar(achar:char):boolean;
function IsPunct(achar:char):boolean;
function IsDigit(achar:char):boolean;
function IsXDigit(achar:char):boolean;
function IsUpper(achar:char):boolean;
function IsLower(achar:char):boolean;
function IsPrint(achar:char):boolean;
function IncludeChar(achar:char;var aset:TAlphabetSet;abit:integer=-1):boolean;
function ExcludeChar(achar:char;var aset:TAlphabetSet;abit:integer=-1):boolean;
function IncludeEnum(pos1,pos2:integer;var aset:TAlphabetSet):boolean;
function ExcludeEnum(pos1,pos2:integer;var aset:TAlphabetSet):boolean;
procedure IncludeSet(var dstset:TAlphabetSet;srcset:TAlphabetSet);
procedure ExcludeSet(var dstset:TAlphabetSet;srcset:TAlphabetSet);
function CharInSet(achar:char;aset:TAlphabetSet):boolean;
function CopySet(dstSet,srcSet:TBits):boolean;
procedure ZeroSet(aset:TBits);

const ALPHA_NONE :TAlphabetSet=($00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00);
const ALPHA_UP   :TAlphabetSet=($00,$00,$00,$00,$00,$00,$00,$00,$7F,$FF,$FF,$E0,$00,$00,$00,$00);
const ALPHA_LO   :TAlphabetSet=($00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$7F,$FF,$FF,$E0);
const ALPHA_CHAR :TAlphabetSet=($00,$00,$00,$00,$00,$00,$00,$00,$7F,$FF,$FF,$E0,$7F,$FF,$FF,$E0);
const ALPHA_DIG  :TAlphabetSet=($00,$00,$00,$00,$00,$00,$FF,$C0,$00,$00,$00,$00,$00,$00,$00,$00);
const ALPHA_XDIG :TAlphabetSet=($00,$00,$00,$00,$00,$00,$FF,$C0,$7E,$00,$00,$00,$7E,$00,$00,$00);
const ALPHA_SPC  :TAlphabetSet=($00,$7C,$00,$00,$80,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00);
const ALPHA_NLN  :TAlphabetSet=($00,$24,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00);
const ALPHA_TAB  :TAlphabetSet=($00,$40,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00);
const ALPHA_PNCT :TAlphabetSet=($FF,$FF,$FF,$FF,$80,$00,$00,$00,$00,$00,$00,$00,$00,$00,$00,$01);
const ALPHA_PRN  :TAlphabetSet=($00,$00,$00,$00,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FE);
const ALPHA_ALL  :TAlphabetSet=($FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF);
implementation

const TKN_LEN    =100;
const MAX_STATES =2047;

const SYM_EPSILON  =-1;
const SYM_CHAR     =-2;
const SYM_SET      =-3;

const S_REJECT    =-1;

const MAT_ALL     =02;
const MAT_CHAR    =04;
const MAT_DIGIT   =08;
const MAT_SEP     =16;
const MAT_HEX     =32;

const A_MAX       =15;
const A_LIT       =42;
const A_CNT       =128;
function  GetChClass(const str:pchar;var aset:TAlphabetSet):integer;forward;
procedure GetEscSet(ch:char;var aset:TAlphabetSet);forward;
destructor TMList.Destroy;
begin
 Clear;
 inherited;
end;

function TMList.InsertOperator(aop:char):boolean;
var
 tmpnode:PTokenNode;
begin
 new(tmpnode);
 tmpnode.data:=nil; tmpnode.tag:=aop;tmpnode.next:=nil;
 if ffront=nil then
  begin
   ffront:=tmpnode;
   frear:=tmpnode;
  end
 else
  begin
   frear.next:=tmpnode;
   frear:=tmpnode;
  end;
 result:=true;
end;

function TMList.InsertMachine(adata:PMachine):boolean;
var
 tmpnode:PTokenNode;
begin
 result:=false;
 if adata=nil then exit;
 new(tmpnode);
 tmpnode.data:=adata; tmpnode.tag:=#0;tmpnode.next:=nil;
 if ffront=nil then
  begin
   ffront:=tmpnode;
   frear:=tmpnode;
  end
 else
  begin
   frear.next:=tmpnode;
   frear:=tmpnode;
  end;
 result:=true;
end;

function TMList.RemoveOperator:char;
var
 tmpNode:PTokenNode; 
begin
 result:=#0;
 if ffront= nil then exit;
 result:=ffront.tag;
 tmpnode:=ffront;
 ffront:=ffront.next;
 dispose(tmpnode);
end;

function TMList.IsMachine:boolean;
begin
 result:=false;
 if ffront=nil then exit;
 if ffront.data=nil then exit;
 result:=true; 
end;

function TMList.RemoveMachine:PMachine;
var
 tmpnode:PTokenNode;
begin
 result:=nil;
 if ffront= nil then exit;
 result:=ffront.data;
 tmpnode:=ffront;
 ffront:=ffront.next;
 dispose(tmpnode);
end;

function TMList.IsEmpty:boolean;
begin
 result:=ffront=nil;
end;

procedure TMList.Clear;
var
 tmpnode:PTokenNode;
begin
 while(ffront<>nil) do
  begin
   tmpnode:=ffront;
   ffront:=ffront.next;
   dispose(tmpnode);
  end;
 frear:=nil;
end;

constructor TLex.Create;
begin
 flist:=TMlist.Create;
 fclosureset:=TBits.Create;
 fbeginset:=TBits.Create;
 fmoveset:=TBits.Create;
 new(fspace);
 new(fstack);
 forigin:=longword(@fspace[0]);
end;

destructor TLex.Destroy;
begin
 dispose(fspace);
 dispose(fmachine);
 dispose(fstack);
 freeandnil(flist);
 freeandnil(fclosureset);
 freeandnil(fmoveset);
 inherited;
end;

function TLex.GetState:PState;
begin
 result:=nil;
 if(fcount=MAX_STATES) then  exit;
 result:=@fspace[fcount];
 zeromemory(result,sizeof(TState));
 result.accept:=S_REJECT;
 fcount:=fcount+1;
end;

function TLex.CreateMachine(ainfo:integer;aset:TAlphabetSet):PMachine;
var
 amachine:PMachine;
begin
 new(amachine);
 amachine.start:=GetState;
 amachine.final:=GetState;
 with (amachine^) do
  begin
   start.cinfo:=ainfo;
   start.ch:=aset;
   start.next1:=amachine.final;
   start.next2:=nil;
   final.next1:=nil;
   final.next2:=nil;
  end;
 result:=amachine;
end;

function TLex.OrMachine(mac1,mac2:PMachine):PMachine;
begin
 result:=nil;
 if(mac1=nil)or(mac2=nil) then exit;
 new(result);
 with(result^) do
  begin
   start:=GetState;
   start.cinfo:=SYM_EPSILON;
   start.next1:=mac1.start;
   start.next2:=mac2.start;
   final:=GetState;
   mac1.final.cinfo:=SYM_EPSILON;
   mac2.final.cinfo:=SYM_EPSILON;
   mac1.final.next1:=final;
   mac2.final.next1:=final;
  end;
 dispose(mac1);
 dispose(mac2);
end;

function TLex.AndMachine(mac1,mac2:PMachine):PMachine;
begin
 result:=nil;
 if(mac1=nil)or(mac2=nil) then exit;
 mac1.final^:=mac2.start^;
 mac1.final:=mac2.final;
 dispose(mac2);
 result:=mac1;
end;

function TLex.KCloseMachine(amac:PMachine):PMachine;
begin
 result:=nil;
 if amac=nil then exit;
 new(result);
 with(result^) do
  begin
   start:=GetState;
   final:=GetState;
   start.cinfo:=SYM_EPSILON;
   start.next1:=amac.start;
   start.next2:=final;
   amac.final.cinfo:=SYM_EPSILON;
   amac.final.next1:=amac.start;
   amac.final.next2:=final;
  end;
 dispose(amac);
end;

function TLex.PCloseMachine(amac:PMachine):PMachine;
begin
 result:=nil;
 if amac=nil then exit;
 new(result);
 with(result^) do
  begin
   start:=GetState;
   final:=GetState;
   start.cinfo:=SYM_EPSILON;
   start.next1:=amac.start;
   amac.final.cinfo:=SYM_EPSILON;
   amac.final.next1:=amac.start;
   amac.final.next2:=final;
  end;
 dispose(amac);
end;

function TLex.QCloseMachine(amac:PMachine):PMachine;
begin
 result:=nil;
 if amac=nil then exit;
 new(result);
 with(result^) do
  begin
   start:=GetState;
   final:=GetState;
   start.cinfo:=SYM_EPSILON;
   start.next1:=amac.start;
   start.next2:=amac.final;
   amac.final.cinfo:=SYM_EPSILON;
   amac.final.next1:=final;
  end;
 dispose(amac);
end;

function TLex.ConstructMachine:PMachine;
var
 tos,tmpnode:PTokenNode;
 res,mac1,mac2:PMachine;
 op:char;
begin
 result:=nil;
 tos:=nil;
 while(not flist.IsEmpty) do
  begin
   if(flist.IsMachine) then
    begin
     new(tmpnode);
     tmpnode.data:=flist.RemoveMachine;
     tmpnode.next:=tos;
     tos:=tmpnode;
    end
   else
    begin
     op:=flist.RemoveOperator;
     if(op in ['.','|']) then
      begin
       if(tos=nil) then exit;
       tmpnode:=tos;mac1:=tos.data;tos:=tos.next;dispose(tmpnode);
       if(tos=nil) then exit;
       tmpnode:=tos;mac2:=tos.data;tos:=tos.next;dispose(tmpnode);
       if(op='.') then res:=AndMachine(mac2,mac1)
       else res:=OrMachine(mac1,mac2);
       new(tmpnode);
       tmpnode.data:=res;
       tmpnode.next:=tos;
       tos:=tmpnode;
      end
     else
      begin
       if(tos=nil) then exit;
       tmpnode:=tos;mac1:=tos.data;tos:=tos.next;dispose(tmpnode);
       if(op='*') then res:=KCloseMachine(mac1)
       else if(op='+')then res:=PCloseMachine(mac1)
       else res:=QCloseMachine(mac1);
       new(tmpnode);
       tmpnode.data:=res;
       tmpnode.next:=tos;
       tos:=tmpnode;
      end;
    end;
  end;
 if(tos=nil) then exit;
 result:=tos.data;
 dispose(tos);
end;

function TLex.GetClosureSet(aset:TBits):boolean;
var
 cur,i:integer;
 curs:PState;
 sidx:integer;
begin
 result:=false; ftos:=-1;zeroset(fclosureset);
 if(aset=nil) then exit;
 for i:=0 to aset.Size-1 do
  begin
   if(aset.Bits[i]) then
    begin
     fclosureset.bits[i]:=true;
     ftos:=ftos+1;
     fstack[ftos]:=i;
    end
  end;
 if(ftos<0) then exit;
 sidx:=MAX_STATES;
 while(ftos>=0) do
  begin
   cur:=fstack[ftos];ftos:=ftos-1;curs:=@fspace[cur];
   if(curs.accept<>S_REJECT) and (cur<sidx) then
    begin
     faccept:=curs;
     sidx:=cur;
    end;
   if (fspace[cur].cinfo=SYM_EPSILON) then
    begin
     if(curs.next1<>nil) then
      begin
       i:=(longword(curs.next1)-forigin) div sizeof(TState);
       fclosureset.Bits[i]:=true;
       ftos:=ftos+1;fstack[ftos]:=i;
      end;
     if(curs.next2<>nil) then
      begin
       i:=(longword(curs.next2)-forigin) div sizeof(TState);
       fclosureset.Bits[i]:=true;
       ftos:=ftos+1;fstack[ftos]:=i;
      end;
    end
  end;
 result:=true;
end;

procedure TLex.GetMoveSet(aset:TBits;ach:char);
 function CharPresent(ach:char;astate:Pstate):boolean;
  begin
   result:=false;
   if(astate.cinfo>=0) and(astate.cinfo=byte(ach)) then result:=true;
   if(astate.cinfo=SYM_SET) and charinset(ach,astate.ch) then result:=true; 
  end;
var
 i,j:integer;
 curs:PState;
begin
 ZeroSet(fmoveset);
 for i:=0 to aset.size-1 do
  begin
   if(aset.Bits[i]=true) then
    begin
     curs:=@fspace[i];
     if CharPresent(ach,curs) then
      begin
       j:=(longword(curs.next1)-forigin) div sizeof(TState);
       fmoveset[j]:=true;
      end;
    end;
  end;
end;

procedure TLex.ClearTokens;
begin
 fcount:=0;fmachine:=nil;ftos:=-1;
 flaccept:=nil;faccept:=nil;
end;
function TLex.ConvertToken(atoken:string):boolean;
 function IsLexOp(achar:char):boolean;
 begin
  result:= achar in ['*','+','?','.','|','(',')'];
 end;
 function prec(ch:char):integer;
 begin
  result:=-1;
  case ch of
   '*':result:=5;
   '+':result:=5;
   '?':result:=5;
   '.':result:=4;
   '|':result:=3;
   '(':result:=2;
  end;
 end;
var
 ainfo,mov,i,j,n:integer;c:char;
 aset:TAlphabetSet;
 prevalpha:boolean;
 stack:pchar;isesc:boolean;
begin
 result:=true; prevalpha:=false;i:=1;j:=-1;
 if(atoken='') then exit; flist.clear;
 atoken:='('+atoken+')';isesc:=false;
 n:=length(atoken);getmem(stack,n*sizeof(char));
 if(n>TKN_LEN) then exit;
 while i<=n do
  begin
   c:=atoken[i];
   if (c='\') then
    begin
     if isesc=false then
      begin
       isesc:=true;i:=i+1;continue;
      end
     else isesc:=false;
    end;
   if isesc or not IsLexOp(c) or (c='[') then
    begin
     if prevalpha then
      begin
       while((j>=0) and (prec(stack[j])>=prec('.'))) do
        begin
         flist.insertOperator(stack[j]);
         j:=j-1;
        end;
       j:=j+1;
       stack[j]:='.';
      end;
     if isesc then
      begin
       aset:=ALPHA_NONE;
       ainfo:=SYM_SET;
       GetEscSet(c,aset);
       flist.InsertMachine(CreateMachine(ainfo,aset));
      end
     else if (c='[') then
      begin
       mov:=GetChClass(@atoken[i],aset);
       if mov=0 then
        begin
         result:=false;break;
        end;
       i:=i+mov;
       ainfo:=SYM_SET;
       flist.InsertMachine(CreateMachine(ainfo,aset));
      end
     else
      begin
       aset:=ALPHA_NONE;
       ainfo:=byte(c);
       flist.InsertMachine(CreateMachine(ainfo,aset));
      end;
     prevalpha:=true;
    end
   else if(c='(') then
    begin
     if prevalpha then
      begin
       while((j>=0) and (prec(stack[j])>=prec('.'))) do
        begin
         flist.insertOperator(stack[j]);
         j:=j-1;
        end;
       j:=j+1;
       stack[j]:='.';
      end;
     prevalpha:=false;
     j:=j+1;stack[j]:=c;
    end
   else if(c in ['*','+','?','|','.']) then
    begin
     if(c in ['|','.']) then prevalpha:=false else prevalpha:=true;
     while((j>=0) and (prec(stack[j])>=prec(c))) do
      begin
       flist.InsertOperator(stack[j]);
       j:=j-1;
      end;
     j:=j+1;
     stack[j]:=c;
    end
   else if(c=')') then
    begin
     prevalpha:=true;
     while((j>=0) and (stack[j]<>'(')) do
      begin
       flist.InsertOperator(stack[j]);
       j:=j-1;
      end;
     j:=j-1;
    end
   else
    begin
     result:=false;
     break;
    end;
   i:=i+1;isesc:=false;
  end;
 freemem(stack);
 if result=false then flist.Clear;
end;

function TLex.AddToken(atoken:string;index:integer;lh:integer=0;eat:boolean=false):boolean;
var
 amac:PMachine;
begin
 result:=ConvertToken(atoken);
 if(result=false) then exit;
 amac:=ConstructMachine;
 if amac=nil then exit;
 amac.final.accept:=index;
 amac.final.lh:=lh;
 amac.final.eat:=eat;  
 if(fmachine=nil) then fmachine:=amac
 else fmachine:=OrMachine(fmachine,amac);
end;

function TLex.ListTokens(aline:string;var tknlist:array of TToken;tkncnt:integer):integer;
var
 st,i,tknidx,n:integer;
 bno,eno:integer;
 flg:boolean;
begin
 result:=0;tknidx:=0;i:=1;bno:=1;eno:=1; 
 if aline='' then exit;
 n:=length(aline);faccept:=nil;flaccept:=nil;
 st:=(longword(fmachine.start)-forigin) div sizeof(TState);
 ZeroSet(fclosureset);ZeroSet(fbeginset);fbeginset.Bits[st]:=true;
 if not GetClosureSet(fbeginset) then exit;
 CopySet(fbeginset,fclosureset);
 while true do
  begin
   GetMoveSet(fclosureset,aline[i]);
   flg:=GetClosureSet(fmoveset);
   if flg then
    begin
     if(faccept<>nil) then
      begin
       eno:=i;
       flaccept:=faccept;
       faccept:=nil;
      end;
     i:=i+1;
    end;
   if (flg=false) or (i>n) then
    begin
     if flaccept=nil then break;
     if(tknidx>=tkncnt) then break;
     tknlist[tknidx].start:=bno;
     tknlist[tknidx].stop:=eno-flaccept.lh;
     tknlist[tknidx].code:=flaccept.accept;
     tknidx:=tknidx+1;
     if not flaccept.eat then i:=i-flaccept.lh;
     flaccept:=nil; CopySet(fclosureset,fbeginset);
     while(i<=n) and (aline[i] in [#0,#9,#32]) do i:=i+1;
     bno:=i;eno:=i;
    end;
   if(i>n) then break;
  end;
 result:=tknidx;
end;

procedure GetEscSet(ch:char;var aset:TAlphabetSet);
begin
 case ch of
  'A':aset:=ALPHA_ALL;
  'C':IncludeSet(aset,ALPHA_UP);
  'c':IncludeSet(aset,ALPHA_LO);
  'a':IncludeSet(aset,ALPHA_CHAR);
  'd':IncludeSet(aset,ALPHA_DIG);
  'x':IncludeSet(aset,ALPHA_XDIG);
  's':IncludeSet(aset,ALPHA_SPC);
  'n':IncludeSet(aset,ALPHA_NLN);
  't':IncludeSet(aset,ALPHA_TAB);
  'p':IncludeSet(aset,ALPHA_PNCT);
  'v':IncludeSet(aset,ALPHA_PRN);
  else IncludeChar(ch,aset);
 end;
end;

function GetChClass(const str:pchar;var aset:TAlphabetSet):integer;
var
 pos1,pos2:integer;
 ibit,cont:boolean;
 lh,c:pchar;
begin
 result:=0;if str=nil then exit;c:=str;
 if c^<>'[' then exit;inc(c);
 if c^='+' then ibit:=true else if c^='-' then ibit:=false else exit;
 inc(c);if c^<>':' then exit;inc(c);cont:=true;lh:=c;
 if ibit then aset:=ALPHA_NONE else aset:=ALPHA_ALL;
 while (cont) do
  begin
   if isAscii(c^) then
    begin
     inc(lh);if lh^=#0 then exit;
     if (lh^='.') then
      begin
       inc(lh);if(lh^<>'.')then exit;
       inc(lh);if lh^=#0 then exit;
       if(lh^='\')then inc(lh);if lh^=#0 then exit;
       if not IsAscii(lh^) then exit;
       pos1:=Ord(c^);pos2:=Ord(lh^);if(pos1>pos2) then exit;
       if ibit then IncludeEnum(pos1,pos2,aset) else ExcludeEnum(pos1,pos2,aset);
       inc(lh);if lh^=#0 then exit;
       if(lh^=']') then
        begin
         result:=1;
         break;
        end
       else if(lh^=',')then
        begin
         inc(lh);c:=lh;
        end
       else exit;
      end
     else if(lh^=',') or (lh^=']')then
      begin
       if ibit then IncludeChar(c^,aset) else ExcludeChar(c^,aset);
       if lh^=']' then
        begin
         result:=1;break;
        end;
       inc(lh);if lh^=#0 then exit;
       c:=lh;
      end
    end
   else exit;
  end;
 if result=1 then result:=longword(lh)-longword(str);
end;

function IsSetEmpty(aset:TAlphabetSet):boolean;
var
 i:integer;
begin
 result:=false;
 for i:=0 to A_MAX do if(aset[i]<>0) then exit;
 result:=true;
end;

function IsAscii(achar:char):boolean;
begin
 result:=ord(achar)<128;
end;

function IsChar(achar:char):boolean;
begin
 result:=achar in ['A'..'Z','a'..'z'];
end;

function IsDigit(achar:char):boolean;
begin
 result:=achar in ['0'..'9'];
end;

function IsXDigit(achar:char):boolean;
begin
 result:=achar in ['0'..'9','A'..'F','a'..'f'];
end;

function IsPunct(achar:char):boolean;
begin
 result:=achar in [#0..#13,#32];
end;

function IsUpper(achar:char):boolean;
begin
 result:=achar in['A'..'Z'];
end;

function IsLower(achar:char):boolean;
begin
 result:=achar in['a'..'z'];
end;

function IsPrint(achar:char):boolean;
begin
 result:=achar in [#32..#126];
end;

function IncludeChar(achar:char;var aset:TAlphabetSet;abit:integer=-1):boolean;
var
 bpos:integer;
 r,c:integer;
begin
 result:=false;
 if abit>=0 then bpos:=abit else bpos:=Ord(achar);
 if (bpos>A_CNT) then exit;
 r:=bpos div 8; c:=7-(bpos mod 8);
 aset[r]:=aset[r] or (1 shl c);
 result:=true;
end;

function ExcludeChar(achar:char;var aset:TAlphabetSet;abit:integer=-1):boolean;
var
 bpos:integer;
 r,c:integer;
begin
 result:=false;
 if abit>=0 then bpos:=abit else bpos:=Ord(achar);
 if(bpos<0) or (bpos>A_CNT) then exit;
 r:=bpos div 8; c:=7-(bpos mod 8);
 aset[r]:=aset[r] and not (1 shl c);
 result:=true;
end;

function IncludeEnum(pos1,pos2:integer;var aset:TAlphabetSet):boolean;
var
 i:integer;
begin
 result:=false;
 for i:=pos1 to pos2 do result:=IncludeChar(#0,aset,i);
end;

function ExcludeEnum(pos1,pos2:integer;var aset:TAlphabetSet):boolean;
var
 i:integer;
begin
 result:=false;
 for i:=pos1 to pos2 do result:=ExcludeChar(#1,aset,i);
end;

procedure IncludeSet(var dstset:TAlphabetSet;srcset:TAlphabetSet);
var
 i:integer;
begin
 for i:= 0 to A_MAX do dstset[i]:=dstset[i] or srcset[i];
end;

procedure ExcludeSet(var dstset:TAlphabetSet;srcset:TAlphabetSet);
var
 i:integer;
begin
 for i:= 0 to A_MAX do dstset[i]:=dstset[i] and not srcset[i];
end;

function CharInSet(achar:char;aset:TAlphabetSet):boolean;
var
 bpos:integer;
 r,c:integer;
begin
 bpos:=Ord(achar);                                    
 r:=bpos div 8;c:=7-(bpos mod 8);
 result:=(aset[r] and (1 shl c))>0;
end;

procedure ZeroSet(aset:TBits);
var
 i:integer;
begin
 for i:=0 to aset.Size-1 do aset.Bits[i]:=false;
end;

function CopySet(dstSet,srcSet:TBits):boolean;
var
 i:integer;
begin
 result:=false;
 if(dstSet=nil) or (srcSet=nil) then exit;
 dstSet.Size:=srcSet.Size;
 for i:=0 to srcSet.Size-1 do
  dstSet.Bits[i]:=srcSet.bits[i];
 result:=true;
end;

end.
